#!/usr/bin/perl
# ONO Hiroki onohiroki@cup.com
use strict;
use warnings;
use lib "./lib";
use Lens;

my $glassCatalog = GlassCatalog->new;
my $glass = Glass->new('BK7','SCHOTT','1.51680003','64.16738393');
$glassCatalog->add_glass($glass);
$glassCatalog->list;
my $wavelength = Wavelength->new->by_names('g','F','d','C');
my $lens = Lens->new('lens');
$lens->set_f_number(4);
$lens->wavelength($wavelength);
$lens->add_surface(0,1E16);
$lens->add_surface(100,10,$glass);
$lens->add_surface(-100,100);
$lens->add_surface(0,0);
$lens->set_indeces;
$lens->list;
$lens->save_dump("examplelens");

use Lens::PlotSVG;
my $h = 6;
my $svg = PlotSVG->new(init_x=>30,init_y=>20,lens_scale=>2);
my @rays = (
	RaySegment->new([1,0,0],[0,0,0],  [1,0,0],0),
	RaySegment->new([1,0,0],[0,$h,0], [1,0,0],0),
	RaySegment->new([1,0,0],[0,-$h,0], [1,0,0],0),
);
foreach my $r (@rays) {
	my $ray = $lens->skew_ray_trace($r);
	$svg->add($ray,'red');
}
$svg->add($lens);
$svg->save_as_file("plot.svg");


