#!/usr/bin/perl
# http://www.opensource.org/licenses/bsd-license
# Copyright (c) 2011, ONO Hiroki onohiroki@cup.com 
# All rights reserved.

{ package Lens;
	use strict;
	use warnings;
	use Lens::Ray;
	use Lens::Glass;
	use Lens::GlassAGF;
	use Lens::Surface;
	use Lens::Wavelength;
	use Lens::CODEV;
	use Lens::ZEMAX;
	#use Lens::LensShell;
	#use Lens::PlotSVG;
	sub license {
		my $self = shift;
		print << 'ENDOFLICENSE';
Perl Module for the lens system.
Open source project.
Copyright (c) 2011, ONO Hiroki
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
Neither the name of the onohiroki.cycling.jp nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
ENDOFLICENSE
		return $self;
	}

	sub new {
		my $class = shift;
		my $name = shift;
		$name = 'untitled lens' unless $name;
		$name =~ s/^\s+//;
		$name =~ s/\s+$//;
		my $surfaces = shift;
		unless ($surfaces) {
			$surfaces = [];
		}
		my $stop_surface = shift;
		unless ($stop_surface) {
			my $num = @$surfaces;
			if (@$surfaces > 1) {
				$stop_surface=1;
			}
			else {
				$stop_surface=0;
			}
		}
		my $number_of_position = 1;
		my $wavelength = Wavelength->wavelength_names('d');
		my $glasscatalog = GlassCatalog->new;
		my $self = {
			Name => $name,
			Surfaces => $surfaces,
			StopSurface => $stop_surface,
			Zoom => {Surfaces => {}},
			CurrentSurface => undef,
			NumberOfPosition => $number_of_position,
			GlassCatalog => $glasscatalog,
			Etc => [],
			Position => {},
			CurrentPosition => 1,
			Wavelength => $wavelength,
			ReferenceWavelength => $wavelength,
			Aperture => {EntrancePupilDiameter => 0},
			Field => [{Angle=>[0,0],Vignetting=>[1,1,1,1]}],
			OpticalAxis => [1,0,0],
		};
		bless $self, $class;
		return $self;
	}

	sub new_with_dumped_data {
		my $class = shift;
		my $dumped_data = shift;
		my $lens;
		my $VAR1;
		my $VAR2;
		my $VAR3;
		my $VAR4;
		my $GLASS_CATALOG_DATA;
		my $ZOOM_DATA;
		my $LENS_DATA;
		if ($dumped_data && ($dumped_data =~ m/VAR/ || $dumped_data =~ m/LENS_DATA/)) {
			eval($dumped_data);
			if ($LENS_DATA){
				$lens = $LENS_DATA;
			}
			elsif ($VAR1 =~ m/Lens\=HASH/ ){
				$lens = $VAR1;
			}
			elsif ($VAR2 =~ m/Lens\=HASH/ ){
				$lens = $VAR2;
			}
			elsif ($VAR3 =~ m/Lens\=HASH/ ){
				$lens = $VAR3;
			}
			elsif ($VAR4 =~ m/Lens\=HASH/ ){
				$lens = $VAR4;
			}
		}
		else {
			my $glasscatalog = GlassCatalog->new;
			$lens = {
				Name => q{},
				Surfaces => [],
				StopSurface => 0,
				Zoom => {Surfaces => {}},
				CurrentSurface => undef,
				NumberOfPosition => 1,
				GlassCatalog => $glasscatalog,
				Etc => [],
				Position => {},
				CurrentPosition => 1,
				Wavelength =>undef,
				ReferenceWavelength => {},
				Aperture => {EntrancePupilDiameter => 0},
				Field => [{Angle=>[0,0],Vignetting=>[1,1,1,1]}],
				OpticalAxis => [1,0,0],
			};
			bless $lens, $class;
		}
		$lens->current_position(1);
		$lens->{Position} = {};
		my $glasscatalog = $lens->reset_glasscatalog;
		my $surfaces = $lens->surfaces;
		$lens->current_surface($lens->last_surface);
		return $lens;
	}

	sub load_yaml_file {
		my $class = shift;
		my $file_name = shift;

		require YAML;
		my $y_lens = YAML::LoadFile("$file_name");
		$y_lens->current_surface($y_lens->last_surface);
		return $y_lens;
	}

	sub new_with_yaml_data { 
		my $class = shift;
		my $data = shift;
		require YAML;
		my $y_lens = YAML::Load($data);
		$y_lens->current_surface($y_lens->last_surface);
		return $y_lens;
	}

	sub new_with_xml_data { 
		my $class = shift;
		my $data = shift;
		require XML::Dumper;
		my $dump = new XML::Dumper;
		my $x_lens = $dump->xml2pl( $data );
		$x_lens->current_surface($x_lens->last_surface);
		return $x_lens;
	}

	sub air {
		my $self = shift;
		my $air = Air->get;
		return $air;
	}

	sub new_with_json_data { 
		my $class = shift;
		my $data = shift;
		require JSON;
		my $json = new JSON;
		my $lens = $json->decode($data)->{Lens};
		my $new_surfaces = [];
		my $surfaces = $lens->{Surfaces};
		my $air = Air->get;
		$lens->{Position}={};
		my $glass_catalog = GlassCatalog->new;
		foreach my $s (@$surfaces) {
			if (exists $s->{Surface}->{Glass}->{Air}){
				$s->{Surface}->{Glass} = $air;
			}
			else {
				my $g = Glass->bless_hash($s->{Surface}->{Glass}->{Glass});
				$s->{Surface}->{Glass} = $glass_catalog->add_glass($g);
			}
			$s = Surface->bless_hash($s->{Surface});
		}
		#use Data::Dumper;
		#print Dumper($glass_catalog);
		$glass_catalog = undef;
		bless $lens, $class;
		$lens->current_surface($lens->last_surface);
		return $lens;
	}

	sub read_file {
		my $self;
		my $file;
		my @glass_catalog_array = ();
		my $cv_glass_catalog;
		my $agf_glass_catalog;
		foreach (@_) {
			if ((ref $_) =~ m/Lens/) {
				$self = $_;
			}
			elsif ((ref $_) eq 'GlassCatalog'){
				push (@glass_catalog_array,$_);
			}
			else {
				$file = $_ if (-e $_); 
			}
		}
		my $gc_num = @glass_catalog_array;
		if (@glass_catalog_array) {
			if (@glass_catalog_array > 1) {
				$cv_glass_catalog = pos @glass_catalog_array;
			}
			$agf_glass_catalog = shift @glass_catalog_array;
		}
		unless ($file) {
			print STDERR "\nread_file\nError: Can't open any file.\n";
			if ((ref $self) =~ m/Lens/) {
				return $self;
			}
			else {
				return Lens->new;
			}
		}
		my $type = q{};
		$type = 'zmx' if ($file && $file =~ m/\.zmx$/i);
		my $lens;
		my $data;
		if (-e $file) {
			open(IN, "<$file");
			while (<IN>) {
				$data = $data.$_;
				$type = 'xml' if (m/\<hashref blessed_package\=\"Lens\"/);	
				$type = 'yaml' if (m/--- !!perl\/hash:Lens/);
				$type = 'json' if (m/"Lens"\s*:\s*\{/);
				$type = 'dumper' if (m/LENS_DATA = bless/);	
				$type = 'dumper' if (m/VAR1 = bless\( \{/);
				$type = 'codev' if (m/RDM\;LEN.*VERSION:.*LENS VERSION:.*Creation Date:/);
			}
			close(IN);
			if ($type eq 'dumper') {
				$lens = Lens->new_with_dumped_data($data);	
			}
			elsif ($type eq 'json') {
				$lens = Lens->new_with_json_data($data);	

			}
			elsif ($type eq 'yaml'){
				$lens = Lens->new_with_yaml_data($data);
			}
			elsif ($type eq 'xml'){
				$lens = Lens->new_with_xml_data($data);
			}
			elsif ($type eq 'codev'){
				$lens = CODEV::read_lens_data_seq($data,$agf_glass_catalog,$cv_glass_catalog);
			}
			elsif ($type eq 'zmx'){
				$lens = ZEMAX::read_lens_data_zmx($data,$agf_glass_catalog,$cv_glass_catalog);
			}
			else {
				print STDERR "ERROR\nUNKOWN File Type.\n";
			}
		}
		else {
			print STDERR "ERROR\nCan not open File $file.\n";
		}
		unless ($lens) {
			if (ref $self) {
				return $self;
			}
			else {
				return Lens->new;
			}
		}
		else {
			$lens->current_surface($lens->last_surface);
		}
		$lens->set_indeces;
		return $lens;
	}

	sub number_of_surfaces {
		my $self = shift;
		my $surfaces = $self->surface;
		my $number_of_surfaces = @$surfaces;
	}

	sub image_surfaces_number {
		my $self = shift;
		my $surfaces = $self->{Surfaces};
		my $number_of_surfaces = @$surfaces;
		return $number_of_surfaces-1;
	}

	sub list {
		my $self = shift;
		my $pos = $self->current_position;
		my $surfaces = $self->surfaces;
		my $number_of_surfaces = @$surfaces;
		if ($number_of_surfaces < 1) {
			print "\nNo data.\n";
			return;
		}
		print "\n",$self->name,"\n";
		my $add_surface = 0;
		while ($number_of_surfaces < 2) {
			my $surfaces = $self->add_surface(0,0);
			$number_of_surfaces = @$surfaces;
			$add_surface++;
		}
		$self->list_surfaces;
		print "\nList of Glasses:";
		my $glasscatalog = $self->glasscatalog;
		$glasscatalog->set_order_by_surfaces($self->surfaces)->list('All Lens Glasses.');
		print "\n";
		$self->print_zoom_params;
		print "\nReference ";
		$self->reference_wavelength->list_in_line;
		print "\n";
		$self->wavelength->list;
		print "\n";
		$self->paraxial_list;
		print "\n";
		while ($add_surface){
			pop @$surfaces;
			$add_surface--;
		}
		return $self;
	}

	sub list_surfaces {
		my $self = shift;
		my $pos = $self->current_position;
		my $surfaces = $self->surfaces;
		my $number_of_surfaces = @$surfaces;
		if ($number_of_surfaces < 1) {
			print "\nNo data.\n";
			return;
		}
		my $stop = $self->stop_surface_number;
		printf "Position: %3d\n", $pos if ($self->number_of_position != 1);
		printf "%s\n", $$surfaces[0]->list('HEADER');
		my $last = @$surfaces;
		my $pre_gla = $$surfaces[0]->glass_name;
		foreach my $sur (@$surfaces) {
			my $switch = q{};
			my $glass = $sur->glass;
			if ( $sur->number == $last-1){
				$switch ='Image';
			}
			elsif ($sur->number == $stop){
				$switch = 'Stop';
			}
			elsif ($sur->number == 0){
				$switch = 'Object';
			}
			elsif ($sur->number == -1){
				$switch = 'Ent.P';
			}
			if ($sur->number == $self->current_surface_number) {
				unless ($switch) {
					$switch = q{*}.$sur->number;
				}
				else {
					$switch = q{*}.$switch;
				}
			}
			if ($sur->glass_name =~ m/^AIR/i && $pre_gla =~ m/^AIR/i ){
				print $sur->list($switch, 'dummy'), "\n";
			}
			else {
				print $sur->list($switch), "\n";
			}
			$pre_gla = $sur->glass_name;
		}
		return $self;
	}

	sub reverse_surfaces {
		my $self = shift;
		my $surfaces = $self->surfaces;

		my @reversed = reverse @$surfaces;
		my @new_surfaces = ();
		my $radius = -$reversed[0]->{Radius};
		my $power = undef;
		my $last_s_num = 0;
		foreach my $s (@reversed) {
			my $new_surface;
			if ($power) {
				$new_surface = Surface->new_with_power(
					$s->{SurfaceNumber}+1,
					$power,
					$s->thickness,
					$s->{Glass},
					$s->{Stop},
					$s->{etc},
					);
			}
			else {
				$new_surface = Surface->new(
					$s->{SurfaceNumber}+1,
					-$radius,
					$s->thickness,
					$s->{Glass},
					$s->{Stop},
					$s->{etc},
					);
			}

			push(@new_surfaces, $new_surface);
			$radius = $s->{Radius};
			if (exists $s->{Power}) {
				$power = $s->{Power};
			}
			else {
				$power = undef;
			}
			$last_s_num = $s->{SurfaceNumber};
		}
		my $new_surface = Surface->new(
			$last_s_num,
			$radius,
			0,
			$$surfaces[0]->{Glass},
			);
		push(@new_surfaces, $new_surface);
		shift @new_surfaces;
		return \@new_surfaces;
	}

	sub paraxial {
		my $self = shift;
		my $seg_obj = RaySegment->new([1,0,0],[ 0, 0.001, 0],[1,0,0],0);
		my $ref_w = $self->reference_wavelength;
		$seg_obj->wavelength($ref_w);
		my $paraxial_data = $self->paraxial_ray_trace($seg_obj)->paraxial;
		my $focal_length = $paraxial_data->{FocalLength};
		my $paraxial_image_distance = $paraxial_data->{ParaxialImageDistance};
		my $num_of_last_seg = $paraxial_data->{NumberOfLastSegment};
		my $beta = 0;
		if ($self->object_surface->thickness < 1e8) {
			$seg_obj = RaySegment->new([1,0,0],[ 0, 0, 0],[1,0.1,0],0);
			$seg_obj->wavelength($ref_w);
			$paraxial_data = $self->paraxial_ray_trace($seg_obj)->paraxial;
			$beta = $paraxial_data->{Beta};
			$paraxial_image_distance = $paraxial_data->{ParaxialImageDistance};
		}
		my $so = $self->object_surface;
		my $si = $self->image_surface;
		my $rear_principal_point = $paraxial_image_distance - $si->index * $focal_length;
		my $reverse_seg_obj = RaySegment->new([1,0,0],[ 0, 1, 0],[1,0,0],$num_of_last_seg);
		$reverse_seg_obj->wavelength($ref_w);
		$paraxial_data = $self->paraxial_ray_trace($reverse_seg_obj,"Reverse")->paraxial;
		my $front_focal_length = $paraxial_data->{FocalLength};
		my $front_paraxial_image_distance = $paraxial_data->{ParaxialImageDistance};
		my $front_principal_point = -($front_paraxial_image_distance - $so->index * $front_focal_length);

		my $seg = RaySegment->new([1,0,0],[ 0, 0, 0],[1,0.1,0],$self->stop_surface_number);
		$seg->wavelength($ref_w);
		my $paraxial_ray = $self->paraxial_ray_trace($seg,"Reverse");
		my $last = $paraxial_ray->number_of_last;
		my $h = $paraxial_ray->get_h_of_segment($last-1);
		my $entrance_pupil_distance = 0;
		if (zero($paraxial_ray->get_u_of_segment($last-1)) != 0){
		 $entrance_pupil_distance = $h / $paraxial_ray->get_u_of_segment($last-1);
		}
		my $u = $paraxial_ray->get_u_of_segment($last-2);
		my $new_seg = RaySegment->new([1,0,0],[ 0, $h, 0],[1,-$u,0],1);
		$new_seg->wavelength($ref_w);
		$paraxial_ray = $self->paraxial_ray_trace($new_seg);
		my $exit_pupil_distance = 0;

		if ( zero($paraxial_ray->last->paraxial_ray_u) != 0 ) {
			$exit_pupil_distance = - $paraxial_ray->last->paraxial_ray_h / $paraxial_ray->last->paraxial_ray_u + $paraxial_image_distance;
		}
		my $surfaces = $self->surfaces;
		my $total_track = 0;
		my $num = 0;
		for my $sur (@$surfaces) {
			next if ($num++ < 1);
			$total_track = $total_track + $sur->thickness;
		}
		if ($so->thickness < 1e8){
			$total_track = $total_track + $so->thickness;
		}
		my $ape = $self->aperture;
		my $entrance_pupil_diameter = $self->entrance_pupil_diameter;
		if (exists $ape->{FNumber}) {
			my $fn = $ape->{FNumber};
			$entrance_pupil_diameter = $focal_length/$fn;
			$self->set_entrance_pupil_diameter($entrance_pupil_diameter);
		}
	
		return {
			EntrancePupilDistance=>$entrance_pupil_distance,
			EntrancePupilDiameter=>$entrance_pupil_diameter,
			ExitPupilDistance=>$exit_pupil_distance,
			ObjectDistance=>$so->thickness,
			FrontPrincipalPoint=>$front_principal_point,
			RearPrincipalPoint=>$rear_principal_point,
			FocalLength=>$focal_length,
			FrontFocalLength=>$front_focal_length,
			Beta=>$beta,
			ImageDistance=>$paraxial_image_distance,
			FrontImageDistance=>$front_paraxial_image_distance,
			TotalTrack=>$total_track,
			
			};
	}

	sub get_entrance_pupil_distance {
		my $self = shift;
		$self->paraxial->{EntrancePupilDistance};
	}
	sub get_exit_pupil_distance {
		my $self = shift;
		$self->paraxial->{ExitPupilDistance};
	}

	sub get_entrance_pupil_surface {
		my $self = shift;
		my $radius = 0;
		$radius = shift if (@_);
		my $glass = $self->object_surface->glass;
		my $epd = $self->get_entrance_pupil_distance;
		my $entrance_pupil_surface = Surface->new("-1",0, - $epd,$glass);
		return $entrance_pupil_surface;
	}
	sub get_exit_pupil_surface {
		my $self = shift;
		my $radius = 0;
		$radius = shift if (@_);
		my $glass = $self->image_surface->glass;
		my $epd = $self->get_exit_pupil_distance;
		my $sur_num = $self->image_surface->number;
		my $exit_pupil_surface = Surface->new($sur_num + 1,0, - $epd,$glass);
		return $exit_pupil_surface;
	}
	sub add_entrance_pupil_surface_into_surfaces {
		my $self = shift;
		my $radius = 0;
		$radius = shift if (@_);
		my $entrance_pupil_surface = $self->get_entrance_pupil_surface($radius);
		my $surfaces = $self->surfaces;
		$entrance_pupil_surface->set_thickness(($entrance_pupil_surface->thickness) - ($self->object_surface->thickness));
		unshift (@$surfaces,$entrance_pupil_surface);
		#$self->list;
		return $self;
	}

	sub add_exit_pupil_surface_into_surfaces {
		my $self = shift;
		my $radius = 0;
		$radius = shift if (@_);
		my $exit_pupil_surface = $self->get_entrance_pupil_surface($radius);
		my $surfaces = $self->surfaces;
		$exit_pupil_surface->set_thickness( - ($exit_pupil_surface->thickness) - ($self->image_surface->thickness) );
		push (@$surfaces,$exit_pupil_surface);
		return $self;
	}

	sub paraxial_ray_trace_with_ep {
		my $self = shift;
		my $ray_segment = shift;
		my $reverse_mode = shift;
		my $radius = 0;
		$radius = shift if (@_);
		my $ray = $self->add_entrance_pupil_surface_into_surfaces($radius)->paraxial_ray_trace($ray_segment,$reverse_mode);
		my $surfaces = $self->surfaces;
		my $entrance_pupil_surface = shift @$surfaces;
		return $ray;
	}

	sub paraxial_ray_trace_with_exit_pupil {
		my $self = shift;
		my $ray_segment = shift;
		my $reverse_mode = shift;
		my $radius = 0;
		$radius = shift if (@_);
		my $ray = $self->add_exit_pupil_surface_into_surfaces($radius)->paraxial_ray_trace($ray_segment,$reverse_mode);
		my $surfaces = $self->surfaces;
		my $entrance_pupil_surface = pop @$surfaces;
		return $ray;
	}

	sub skew_ray_trace_with_ep {
		my $self = shift;
		my $ray_segment = shift;
		my $reverse_mode = shift;
		my $radius = 0;
		$radius = shift if (@_);
		my $surfaces = $self->surfaces;
		$$surfaces[0]->set_thickness($self->get_entrance_pupil_distance) if ($self->object_surface->thickness > 1e10);
		my $ray = $self->add_entrance_pupil_surface_into_surfaces($radius)->skew_ray_trace($ray_segment,$reverse_mode);
		my $entrance_pupil_surface = shift @$surfaces;
		return $ray;
	}

	sub skew_ray_trace_with_exit_pupil {
		my $self = shift;
		my $ray_segment = shift;
		my $reverse_mode = shift;
		my $radius = 0;
		$radius = shift if (@_);
		my $surfaces = $self->surfaces;
		my $ray = $self->add_exit_pupil_surface_into_surfaces($radius)->skew_ray_trace($ray_segment,$reverse_mode);
		my $entrance_pupil_surface = shift @$surfaces;
		return $ray;
	}


	sub paraxial_list {
		my $self = shift;
		$self->list_zoomed_paraxial;
	}

	sub list_zoomed_paraxial {
		my $self = shift;
		my $num_of_positions = $self->number_of_position;
		my $data = [];
		my $format = " %24s ";
		printf $format, q{};
		my $backup_current_position = $self->current_position;
		foreach my $z (1..$num_of_positions){
			$self->current_position($z);
			my $p = $self->paraxial;
			push (@$data, $p);
			printf " %11s ", "Position:$z" if ($num_of_positions > 1);
		}
		$self->current_position($backup_current_position);
		print "\n";
		foreach my $k (
			'FocalLength',
			#'FrontFocalLength',
			'Beta',
			'FrontPrincipalPoint',
			'RearPrincipalPoint',
			'EntrancePupilDistance',
			'ExitPupilDistance',
			'ObjectDistance',
			'ImageDistance',
			'FrontImageDistance',
			'TotalTrack',
			) {
			#print "\n" if ($k eq 'FocalLength');
			my $str = $k;
			$str =~ s/([a-z])([A-Z])/$1 $2/g;
			printf $format, $str;
			foreach my $p (@$data){
				my $value = $$p{$k};
				if ($value > 1e8) {
					printf " %11g ", $value;
				}
				else {
					printf " %11.5f ", $value;
				}
			}
			print "\n";
		}
		return $self;
	}

	sub paraxial_ray_trace_ {
		my $self = shift;
		my $ray_segment = shift;
		my $wavelength = $ray_segment->wavelength;
		if ((ref $wavelength) ne 'Wavelength') {
			$wavelength = $self->reference_wavelength;
		}
		my $start_sur_num = $ray_segment->num_of_surface;
		my $surfaces = $self->surfaces;

		my $reverse_mode = shift;
		if ($reverse_mode && ($reverse_mode =~ m/R/i || $reverse_mode =~ m/Y/i ) ) {
			$surfaces = $self->reverse_surfaces;
			$start_sur_num = $start_sur_num -1;
		}
		else {
			$reverse_mode = undef;
		}
		my $inv = $self->surface($start_sur_num)->index($wavelength); # 2.11 N
		if ($start_sur_num != 0){
			$inv = $self->surface($start_sur_num-1)->index($wavelength);
		}
		my $ray = Ray->new;
		unless ($inv) {
			print STDERR "\nERROR:paraxial_ray_trace: Not defined index of ",$ray_segment->wavelength_name,"\n"; 
			$ray->add_segment($ray_segment);
			return $ray;
		}
		my $h = @{$ray_segment->get_t}[1];
		my $u = @{$ray_segment->get_q}[1];
		$h = zero($h);
		$u = - zero($u);
		
		my $p = $self->surface($start_sur_num)->power($inv, $wavelength); #2.11
		$inv = $self->surface($start_sur_num)->index($wavelength); #2.11
		my $e = $self->surface($start_sur_num)->thickness / $inv; #2.11
		my $a = $inv * $u; #2.11
		$ray_segment->paraxial_ray_a($a);
		$ray->add_segment($ray_segment);
		my $s;
		foreach my $sur (@$surfaces){
			unless ($reverse_mode) {
				next if ($sur->number < $start_sur_num + 1);
			}
			else {
				next if ($sur->number > $start_sur_num );
			}
			$h = $h - $e * $a;
			$h = zero($h);
			$p = $sur->power($inv, $wavelength);
			$inv = $sur->index($wavelength);
			unless ($inv) {
				if ($ray_segment->wavelength_name) {
					print STDERR "\nERROR:paraxial_ray_trace: Not defined index of ", $ray_segment->wavelength_name," at Surface:", $sur->number, ",\n"; 
				}
				else {
					print STDERR "\nERROR:paraxial_ray_trace: Not defined wavelength:(RaySegmet->wavelength_name) at Surface:", $sur->number, ",\n"; 
				}
				$ray->add_segment($ray_segment);
				return $ray;
			}
			$e = $sur->thickness / $inv;
			$a = $a + $h * $p;
			$a = zero($a);
			$u = $a / $inv;
			if ($a != 0) {
				$s = $inv * $h / $a;
			}
			else {
				$s = 1e15;
			}
			$ray_segment = RaySegment->new([1,0,0],[0,$h,0],[1,-$u,0],$sur->number,$sur->thickness);
			$ray_segment->wavelength($wavelength);
			$ray_segment->paraxial_ray_a($a);
			$ray->add_segment($ray_segment);
		}
		return $ray;
	}

	sub paraxial_ray_trace {
		my $self = shift;
		my $ray_segment = shift;
		my $wavelength = $ray_segment->wavelength;
		if ((ref $wavelength) ne 'Wavelength') {
			$wavelength = $self->reference_wavelength;
		}
		my $start_sur_num = $ray_segment->num_of_surface;
		my $surfaces = $self->surfaces;

		my $reverse_mode = shift;
		if ($reverse_mode && ($reverse_mode =~ m/R/i || $reverse_mode =~ m/Y/i ) ) {
			$surfaces = $self->reverse_surfaces;
			$start_sur_num = $start_sur_num -1;
		}
		else {
			$reverse_mode = undef;
		}
		my $inv = $self->surface($start_sur_num)->index($wavelength);# 2.11 N
		if ($start_sur_num > 0){
			$inv = $self->surface($start_sur_num-1)->index($wavelength);
		}
		my $ray = Ray->new;
		unless ($inv) {
			print STDERR "\nERROR:paraxial_ray_trace: Not defined index of ",$ray_segment->wavelength_name,"\n"; 
			$ray->add_segment($ray_segment);
			return $ray;
		}
		my $h = @{$ray_segment->get_t}[1];
		my $u = @{$ray_segment->get_q}[1];
		$h = zero($h);
		$u = - zero($u); 
		my $p = $self->surface($start_sur_num)->power($inv, $wavelength); #2.11
		$inv = $self->surface($start_sur_num)->index($wavelength); # 2.11 N'
		my $e = $self->surface($start_sur_num)->thickness / $inv; #2.11
		my $a = $inv * $u; #2.11
		$ray_segment->paraxial_ray_a($a);
		my $d = $self->surface($start_sur_num)->thickness;
		$d = -$d if ($reverse_mode);
		$ray_segment->thickness($d);
		$ray->add_segment($ray_segment);
		my $s;
		foreach my $sur (@$surfaces){
			unless ($reverse_mode) {
				next if ($sur->number <= $start_sur_num );
			}
			else {
				next if ($sur->number > $start_sur_num );
			}
			$h = $h - $e * $a;
			$h = zero($h);
			$p = $sur->power($inv, $wavelength);
			$inv = $sur->index($wavelength);
			unless ($inv) {
				if ($ray_segment->wavelength_name) {
					print STDERR "\nERROR:",(ref $self),"::paraxial_ray_trace : Not defined index of ", $ray_segment->wavelength_name," at Surface:", $sur->number, ",\n"; 
				}
				else {
					print STDERR "\nERROR:",(ref $self),"::paraxial_ray_trace : Not defined wavelength:(RaySegmet->wavelength_name) at Surface:", $sur->number, ",\n"; 
				}
				$ray->add_segment($ray_segment);
				return $ray;
			}
			$e = $sur->thickness / $inv;
			$a = $a + $h * $p;
			$a = zero($a);
			$u = $a / $inv;
			if ($a != 0) {
				$s = $inv * $h / $a;
			}
			else {
				$s = 1e15;
			}
			my $d = $sur->thickness;
			$d = -$d if ($reverse_mode);
			$ray_segment = RaySegment->new([1,0,0],[0,$h,0],[1,-$u,0],$sur->number,$d);
			$ray_segment->wavelength($wavelength);
			$ray_segment->paraxial_ray_a($a);
			$ray->add_segment($ray_segment);
		}
		return $ray;
	}

	sub paraxial_ray_list {
		my $self = shift;
		my $ray_segment;
		if (@_) {
			$ray_segment = shift;
		}
		else {
			$ray_segment = RaySegment->new([1,0,0],[ 0,1,0],[1,0,0],0);
		}
		my $ray = $self->paraxial_ray_trace($ray_segment);
		$ray->list;
		return $ray;
	}

	sub power_list {
		my $self = shift;
		my $surfaces = $self->surfaces($self->current_position);
		my $index = 1;
		my $sum = 0;
		print "Power of Surface:\n";
		printf "%7s %15s %10s %10s %15s\n", "Surface#", "Power", "Thickness", "Index", "(Thi. x Ind.)";
		foreach my $sur (@$surfaces) {
			my $sur_mark = ")";
			my $power = $sur->power($index);
			my $thi = $sur->thickness;
			$sum = $sum + $power;
			$index = $sur->index;
			$sur_mark = ">" if ($sur->{Stop});
			if ($thi > 1e10 || $thi < -1e10) {
				printf "%7d%1s %15.10f %10.4g %10.4f %15.8g\n", $sur->number, $sur_mark,$power,$thi,$index,$thi*$index;
			}
			else {
				printf "%7d%1s %15.10f %10.4f %10.4f %15.8f\n", $sur->number, $sur_mark,$power,$thi,$index,$thi*$index;
			}
		}
		#printf "\nSUM of Surface Power:%10.4f\n",$sum;
		#$self->paraxial("list");
		return $sum;
	}

	sub skew_ray_trace {
		my $self = shift;
		my $ray_segment = shift;
		my $wavelength = $ray_segment->wavelength;
		if ((ref $wavelength) ne 'Wavelength') {
			$wavelength = $self->reference_wavelength;
		}
		my $start_sur_num = $ray_segment->num_of_surface;
		my $entrance_pupil = 0;
		$entrance_pupil = 1 if ($start_sur_num == -1);
		my $surfaces = $self->surfaces;
		my $reverse_mode = shift;
		if (@_){
			$$surfaces[0] = shift;
			print "\nSkew Ray Trace with EP Surface:\n";
			$self->list;
		}

		if ($reverse_mode && ($reverse_mode =~ m/R/i || $reverse_mode =~ m/Y/i ) ) {
			$surfaces = $self->reverse_surfaces;
			$start_sur_num = $start_sur_num -1;
		}
		else {
			$reverse_mode = undef;
		}
		my $inv = $self->surface($start_sur_num)->index($wavelength);
		my $ray = Ray->new;
		unless ($inv) {
			print STDERR "\nERROR: Not defined index of ",$ray_segment->wavelength_name,".\n"; 
			$ray->add_segment($ray_segment);
			return $ray;
		}
		my $d = $self->surface($start_sur_num)->thickness;
		my $thickness = $d;
		$thickness = - $thickness if ($reverse_mode);
		$ray_segment->thickness($thickness);
		my ($x, $y, $z) = @{$ray_segment->get_t};
		my ($a, $b, $c) = @{$ray_segment->get_q};
		$a = zero($a);
		$b = zero($b);
		$c = zero($c);
		$x = zero($x);
		$y = zero($y);
		$z = zero($z);
		$b = -$b if ($d < 0);
		if ($a == 0 && $b == 0 && $c == 0) {
			print STDERR "\nERROR: an illegal vector of ray segment.\n"; 
			$ray->add_segment($ray_segment);
			return $ray;
		}
		my $optical_axis = $ray_segment->optical_axis;

		my ($a1, $b1, $c1) = ($a, $b, $c);
		$a1 = $a/$a if ($a != 0);
		$b1 = $b/$b if ($b != 0);
		$c1 = $c/$c if ($c != 0);
		$d = 0 if ($$optical_axis[0] == $a1 && $$optical_axis[1] == $b1 && $$optical_axis[2] == $c1);
		if ($$optical_axis[0] == 0 && $$optical_axis[1] == 0 && $$optical_axis[2] == 1) {
			($x, $y, $z) = ($z, $y, $x);
			($a, $b, $c) = ($c, $b, $a);
		}

		$ray->add_segment($ray_segment);
		unless ($reverse_mode) {
			$start_sur_num = $start_sur_num + 1
		}
		foreach my $sur (@$surfaces){
			#print "\nroop:",$sur->list;
			if ($reverse_mode) {
				next if ($sur->number > $start_sur_num );
			}
			else {
				next if ($sur->number < $start_sur_num );
			}
			$d = 0 if ($sur->number == 1 && $entrance_pupil && $self->object_surface->thickness > 1e10);
			my $intersection_data = $sur->get_intersection([$x, $y, $z],[$a, $b, $c],$d);
			unless ($intersection_data){
				$ray->add_segment($ray_segment);
				return $ray;
			}
			($x, $y, $z) = @{$$intersection_data[0]};
			my $direction =  $sur->get_direction_cosines($inv,$intersection_data,$wavelength);
			($a, $b, $c) = @$direction;
			$a = zero($a);
			$b = zero($b);
			$c = zero($c);
			$x = zero($x);
			$y = zero($y);
			$z = zero($z);
			$d = $sur->thickness;
			$thickness = $d;
			$thickness = - $thickness if ($reverse_mode);
			if ($sur->reflect && $d < 0) {
				$b = -$b;
				$c = -$c;
			}
			my $new_segment;
			$new_segment = RaySegment->new($optical_axis,[$x, $y, $z],[$a, $b, $c],$sur->number,$thickness);
			$new_segment->wavelength($wavelength);
			$ray->add_segment($new_segment);
			$inv = $sur->index($wavelength);
			unless ($inv) {
				print STDERR "\nERROR: Not defined index of ", $ray_segment->wavelength_name," at Surface:", $sur->number, ",\n"; 
				$ray->add_segment($ray_segment);
				return $ray;
			}
		}
		return $ray;
	}

	sub skew_ray_list {
		my $self = shift;
		my $ray = $self->skew_ray_trace(@_);
		$ray->list;
		return $ray;
	}

	sub current_position {
		my $self = shift;
		my $current_position = 1;
		if (@_) {
			$self->{CurrentPosition} = shift;
			return $self;
		}
		unless (exists $self->{CurrentPosition}) {
			return 1;
		}
		else {
			return $self->{CurrentPosition};
		}
	}

	sub set_current_position {
		my $self = shift;
		$self->{CurrentPosition} = shift;
		return $self;
	}

	sub position {
		my $self = shift;
		if (@_) {
			$self->reset_position;
			$self->{Position} = shift;
		}
		$self->{Position};
	}

	sub reset_position {
		my $self = shift;
		$self->{Position} = {};
		return $self;
	}

	sub surfaces {
		my $self = shift;
		my $position = $self->current_position;
		$position  = shift if (@_);
		my $surfaces = $self->{Surfaces};
		return $surfaces if (@$surfaces < 1);
		my $pos_data = $self->position;
		$pos_data->{1} = $surfaces unless (exists $pos_data->{1});
		unless (exists $pos_data->{$position}) {
			my $zoom_data = $self->zoom->{Surfaces};
			my $new_p_s = [];
			foreach my $s (@$surfaces){
				my $n = $s->number;
				my $thickness = $s->thickness;
				if (exists $zoom_data->{$n} && exists $zoom_data->{$n}->{Thickness}) {
					my $arr = $zoom_data->{$n}->{Thickness};
					$thickness = $$arr[$position -1];
				}
				my $glass;
				if ($s->glass) {
					$glass = $s->glass;
				}
				else {
					print STDERR "ERROR:GLASS\n";
					$glass = Air->get;
				}
				my $new_s = Surface->new(
					$n,
					$s->radius,
					$thickness,
					$glass,
					$s->stop,
					$s->number_of_position,
					$s->etc,
					);
				$new_s->list;
				push(@$new_p_s, $new_s);
			}
			$pos_data->{$position} = $new_p_s;
		}
		return $pos_data->{$position};
	}

	sub renumber_surfaces {
		my $self = shift;
		my $surfaces = $self->{Surfaces};
		my $num = 0;
		foreach my $sur (@$surfaces){
			$self->stop_surface_number($num) if ($sur->stop);
			$sur->number($num++);
		}
		$self->reset_position;
		return $self->surfaces;
	}

	sub surface {
		my $self = shift;
		my $sur_num;
		if (@_) {
			$sur_num = shift;
		}
		else {
			$sur_num = $self->current_surface_number;
		}
		$sur_num = 0 if ($sur_num < 0);
		my $surfaces = $self->surfaces($self->current_position);
		if ($sur_num > $#{$surfaces}){
			return $$surfaces[$#{$surfaces}];
		}
		else {
			return $$surfaces[$sur_num];
		}
	}

	sub replace_surface {
		my $self = shift;
		my $sur_num = shift;
		my $old_sur = $self->surface($sur_num);
		my $surfaces = $self->{Surfaces};
		if ((ref $_[0]) =~ m/Surface/) {
			$$surfaces[$sur_num] = shift;
		}
		else {
			my $sur_num_B = shift;
			my $surface_B = $self->surface($sur_num_B);
			$$surfaces[$sur_num_B] = $old_sur;
			$$surfaces[$sur_num] = $surface_B;
		}
		return $self;
	}

	sub current_surface {
		my $self = shift;
		if (@_) {
			$self->{CurrentSurface} = shift;
		}
		my $csn = $self->{CurrentSurface};
		if ($csn && (ref $csn) !~ m/Surface/) {
			print STDERR "\nERROR in ",(ref $self),"::current_surface. {CurrentSurface} $csn is not object.";
			use Data::Dumper;
			print STDERR Dumper($csn);
		}
		return $csn;
	}

	sub previous_surface {
		my $self = shift;
		$self->surface($self->current_surface_number -1);
	}

	sub next_surface {
		my $self = shift;
		$self->surface($self->current_surface_number +1);
	}

	sub current_surface_number {
		my $self = shift;
		my $cs = $self->current_surface;
		if (@_) {
			my $c = shift;
			my $num = 0;
			if ((ref $c) && (ref $c) =~ m/Surface/) {
				$cs = $self->current_surface($c);
			}
			else {
				$cs = $self->current_surface($self->surface($c));
			}
		}
		if ($cs) {
			return $cs->number;
		}
		else {
			return 0;
		}
	}

	sub object_surface {
		my $self = shift;
		$self->surface(0);
	}

	sub image_surface {
		my $self = shift;
		$self->last_surface;
	}

	sub last_surface {
		my $self = shift;
		my $surfaces = $self->surfaces;
		$$surfaces[$#{$surfaces}];
	}

	sub add_surface {
		my $self = shift;
		my $surfaces = $self->surfaces;
		my $num_of_surfaces = @$surfaces;
		my $current_surface = $self->current_surface_number;
		my $new_surface;
		if ((ref $_[0]) =~ m/Surface/) {
			$new_surface = shift;
		}
		else {
			$new_surface = Surface->new($num_of_surfaces,@_);
			$new_surface->glass($self->air) unless ($new_surface->glass);
		}
		if ($num_of_surfaces == 1 && $self->{StopSurface} == 0){
			$self->{StopSurface} = 1;
			$new_surface->set_stop('stop');
		}
		#push (@$surfaces, $new_surface);
		$current_surface++;
		if ($current_surface>$num_of_surfaces) {
			push (@$surfaces, $new_surface);
		}
		else{
			splice(@$surfaces,$current_surface,0,$new_surface);
		}
		$self->renumber_surfaces;
		$self->current_surface($new_surface);
		return $surfaces;
	}

	sub delete_surface {
		my $self = shift;
		my $surfaces = $self->{Surfaces};
		my $num_of_surfaces = @$surfaces;
		my $current_surface = $self->current_surface_number;
		my $num_of_del_sur = $current_surface;
		$num_of_del_sur = shift;
		return $surfaces if ($num_of_del_sur < 0);
		$num_of_del_sur = $num_of_surfaces if ($num_of_del_sur > $num_of_surfaces);
		my $deleted_surface = splice(@$surfaces,$current_surface,1);
		$self->renumber_surfaces;
		return $deleted_surface;
	}

	sub delete_surfaces {
		my $self = shift;
		my $surfaces = $self->{Surfaces};
		foreach my $snum (@_) {
			@$surfaces = grep {$_->number != $snum} @$surfaces;
		}
		return $self->renumber_surfaces;
	}

	sub add_surface_with_power {
		my $self = shift;
		my $surfaces = $self->surfaces;
		my $num_of_surfaces = @$surfaces;
		my $current_surface = $self->current_surface_number;
		my $new_surface;
		my $num_of_args = @_;
		if ($num_of_args < 3) {
			$new_surface = Surface->new_with_power($num_of_surfaces,@_,$self->air) ;
		}
		else {
			$new_surface = Surface->new_with_power($num_of_surfaces,@_);
		}
		if ($num_of_surfaces == 1 && $self->{StopSurface} == 0){
			$self->{StopSurface} = 1;
			$new_surface->set_stop('stop');
		}
		$current_surface++;
		if ($current_surface>$num_of_surfaces) {
			push (@$surfaces, $new_surface);
		}
		else{
			splice(@$surfaces,$current_surface,0,$new_surface);
		}
		$self->renumber_surfaces;
		$self->current_surface($new_surface);
		return $surfaces;
	}

	sub stop_surface {
		my $self = shift;
		$self->surfaces($self->stop_surface_number);
	}

	sub set_stop_surface {
		my $self = shift;
		my $sur_num = shift;
		$self->surface($self->stop_surface_number)->unset_stop;
		$self->surface($sur_num)->set_stop('stop');
		$self->{StopSurface} = $sur_num;
		return $self;
	}

	sub stop_surface_number {
		my $self = shift;
		if (@_) {
			$self->{StopSurface} = shift;
		}
		return $self->{StopSurface};
	}

	sub aperture {
		my $self = shift;
		if (@_) {
			$self->{Aperture} = shift;
		}
		return $self->{Aperture};
	}

	sub reset_aperture {
		my $self = shift;
		if (@_) {
			$self->{Aperture} = shift;
		}
		else {
			my $epd = $self->{Aperture}->{EntrancePupilDiameter};
			$self->{Aperture} = {EntrancePupilDiameter => $epd};
		}
		return $self;
	}

	sub entrance_pupil_diameter {
		my $self = shift;
		my $aperture = $self->aperture; 
		if (@_) {
			my $entrance_pupil = shift;
			$$aperture{EntrancePupilDiameter} = $entrance_pupil;
		}
		return $$aperture{EntrancePupilDiameter};
	}

	sub set_entrance_pupil_diameter {
		my $self = shift;
		my $entrance_pupil = shift;
		my $aperture = $self->aperture; 
		$$aperture{EntrancePupilDiameter} = $entrance_pupil;
		return $self;
	}

	sub f_number {
		my $self = shift;
		my $aperture = $self->aperture; 
		if (@_) {
			my $f_number = shift;
			$$aperture{FNumber} = $f_number;
		}
		return $$aperture{FNumber};
	}

	sub set_f_number {
		my $self = shift;
		my $f_number = shift;
		my $aperture = $self->aperture; 
		$$aperture{FNumber} = $f_number;
		return $self;
	}

	sub field {
		my $self = shift;
		if (@_) {
			$self->{Field} = shift;
		}
		return $self->{Field};
	}

	sub field_angle {
		my $self = shift;
		my $field = $self->field; 
		while (@_){
			my $angle = shift;
			push(@$field, {Angle => $angle,Vignetting => [1,1,1,1]});
		}
		my $angle = [];
		foreach (@$field) {
			push (@$angle,$$_{Angle})
		}
		return $angle;
	}

	sub set_y_field_angle {
		my $self = shift;
		my $field = $self->field;
		my $new_angle = [];
		while (@_) {
			my $angle = shift;
			my $f = shift @$field;
			my $s = $$f{Angle}[1];
			$s = 0 unless ($s);
			push(@$new_angle, {Angle => [$angle,$s], Vignetting => [1,1,1,1]});
		}
		$self->field($new_angle);
		return $self;
	}

	sub set_s_field_angle {
		my $self = shift;
		my $field = $self->field;
		my $new_angle = [];
		while (@_) {
			my $angle = shift;
			my $f = shift @$field;
			my $y = $$f{Angle}[0];
			$y = 0 unless ($y);
			push(@$new_angle, {Angle => [$y,$angle], Vignetting => [1,1,1,1]});
		}
		$self->field($new_angle);
		return $self;
	}

	sub set_y_field_paraxial_image_height {
		my $self = shift;
		my $field = $self->field;
		my $new_height = [];
		while (@_) {
			my $height = shift;
			my $f = shift @$field;
			my $s = $$f{ParaxialImageHeight}[1];
			$s = 0 unless ($s);
			push(@$new_height, {ParaxialImageHeight => [$height,$s], Vignetting => [1,1,1,1]});
		}
		$self->field($new_height);
		return $self;
	}
	sub set_s_field_paraxial_image_height {
		my $self = shift;
		my $field = $self->field;
		my $new_height = [];
		while (@_) {
			my $height = shift;
			my $f = shift @$field;
			my $y = $$f{ParaxialImageHeight}[0];
			$y = 0 unless ($y);
			push(@$new_height, {ParaxialImageHeight => [$y,$height], Vignetting => [1,1,1,1]});
		}
		$self->field($new_height);
		return $self;
	}
	sub set_y_field_image_height {
		my $self = shift;
		my $field = $self->field;
		my $new_height = [];
		while (@_) {
			my $height = shift;
			my $f = shift @$field;
			my $s = $$f{ImageHeight}[1];
			$s = 0 unless ($s);
			push(@$new_height, {ImageHeight => [$height,$s], Vignetting => [1,1,1,1]});
		}
		$self->field($new_height);
		return $self;
	}
	sub set_s_field_image_height {
		my $self = shift;
		my $field = $self->field;
		my $new_height = [];
		while (@_) {
			my $height = shift;
			my $f = shift @$field;
			my $y = $$f{ImageHeight}[0];
			$y = 0 unless ($y);
			push(@$new_height, {ImageHeight => [$y,$height], Vignetting => [1,1,1,1]});
		}
		$self->field($new_height);
		return $self;
	}
	sub set_y_field_object_height {
		my $self = shift;
		my $field = $self->field;
		my $new_height = [];
		while (@_) {
			my $height = shift;
			my $f = shift @$field;
			my $s = $$f{ObjectHeight}[1];
			$s = 0 unless ($s);
			push(@$new_height, {ObjectHeight => [$height,$s], Vignetting => [1,1,1,1]});
		}
		$self->field($new_height);
		return $self;
	}
	sub set_s_field_object_height {
		my $self = shift;
		my $field = $self->field;
		my $new_height = [];
		while (@_) {
			my $height = shift;
			my $f = shift @$field;
			my $y = $$f{ObjectHeight}[0];
			$y = 0 unless ($y);
			push(@$new_height, {ObjectHeight => [$y,$height], Vignetting => [1,1,1,1]});
		}
		$self->field($new_height);
		return $self;
	}
	sub real_image_height_to_angle {
		my $self = shift;
		my $fields = $self->field;
		my $num = $self->last_surface->number;
		my $pi = atan2(1,1) * 4;
		my $deg2rad = 1 / 360 * 2 * $pi;
		my $rad2deg = 1 / $deg2rad;
		foreach my $f (@$fields) {
			my $rayseg = RaySegment->new([1,0,0],[ 0,$f->{ParaxialImageHeight}[0],$f->{ParaxialImageHeight}[1]],[1,0,0],$num);
			my $u1 = $self->paraxial_ray_trace_with_exit_pupil($rayseg,'R')->last->get_q->[1];
			my $u2 = $self->paraxial_ray_trace_with_exit_pupil($rayseg,'R')->last->get_q->[2];
			$f->{Angle}=[$u1 * $rad2deg, $u2 * $rad2deg];
		}
		return $self;
	}
	sub image_height_to_angle {
		my $self = shift;
		my $fields = $self->field;
		my $num = $self->last_surface->number;
		my $pi = atan2(1,1) * 4;
		my $deg2rad = 1 / 360 * 2 * $pi;
		my $rad2deg = 1 / $deg2rad;
		foreach my $f (@$fields) {
			my $rayseg = RaySegment->new([1,0,0],[ 0,$f->{ImageHeight}[0],$f->{ImageHeight}[1]],[1,0,0],$num);
			my $u1 = $self->skew_ray_trace_with_exit_pupil($rayseg,'R')->last->get_q->[1];
			my $u2 = $self->skew_ray_trace_with_exit_pupil($rayseg,'R')->last->get_q->[2];
			$f->{Angle}=[$u1 * $rad2deg, $u2 * $rad2deg];
		}
		return $self;
	}
	sub set_vignetting1 {
		my $self = shift;
		$self->set_vignetting_with_number(1,@_);
		return $self;
	}

	sub set_vignetting2 {
		my $self = shift;
		$self->set_vignetting_with_number(2,@_);
		return $self;
	}

	sub set_vignetting3 {
		my $self = shift;
		$self->set_vignetting_with_number(3,@_);
		return $self;
	}

	sub set_vignetting4 {
		my $self = shift;
		$self->set_vignetting_with_number(4,@_);
		return $self;
	}

	sub set_vignetting_with_number {
		my $self = shift;
		my $num = shift;
		my $field = $self->field;
		my $new_angle = [];
		while (@_) {
			my $v = shift;
			my $f = shift @$field;
			my $vig = $$f{Vignetting};
			$$vig[$num - 1] = $v;
			push(@$new_angle, {%$f, Vignetting => $vig});
		}
		$self->field($new_angle);
		return $self;
	}

	sub vignetting {
		my $self = shift;
		my $fields = $self->field; 
		if (@_) {
			my $new_angle = [];
			while (@_){
				my $v = shift;
				my $f = shift @$fields;
				push(@$new_angle, {%$f, Vignetting => $v});
			}
			$self->field($new_angle);
		}
		my $vig = [];
		foreach (@$) {
			push (@$vig,$$_{Vignetting})
		}
		return $vig;
	}

	sub get_vignetting {
		my $self = shift;
		my $num = shift;
		my $field = $self->vignetting;
		my @vig = map {$$_[$num-1]} @$field;
		return \@vig;
	}

	sub zoom {
		my $self = shift;
		$self->{Zoom} = shift if (@_);
		return $self->{Zoom};
	}

	sub number_of_position {
		my $self = shift;
		$self->{NumberOfPosition} = shift if (@_);
		$self->{NumberOfPosition};
	}

	sub set_number_of_position {
		my $self = shift;
		my $num = shift;
		$self->{NumberOfPosition} = $num;
	}

	sub set_number_of_position_from_thickness {
		my $self = shift;
		my $surface_num = shift;
		if ($self->surface($surface_num)->number_of_position > $self->number_of_position) {
			$self->set_number_of_position($self->surface($surface_num)->number_of_position);
		}
		#$self->surface($surface_num);
		return $self;
	}

	sub set_surface_zoom_data{
		my $self = shift;
		my $surface_num = shift;
		my $kind = shift;
		my $params = shift;
		my %zoom_thickness = ($kind => $params);
		my $surfaces=$self->zoom->{Surfaces};
		$$surfaces{$surface_num} = \%zoom_thickness;
	}

	sub set_zoomed_thickness_of_surface {
		my $self = shift;
		my $surface_num = shift;
		my $params = \@_;
		my $num_of_params = @$params;
		my $zooed_thickness = $self->surface($surface_num)->set_zoomed_thickness(@$params);
		$self->set_surface_zoom_data($surface_num,'Thickness',$zooed_thickness);
		$self->set_number_of_position_from_thickness($surface_num);
		return $self;
	}

	sub set_thickness_with_position_of_surface {
		my $self = shift;
		my $surface_num = shift;
		my $thickness = shift;
		my $position = shift;
		my $zooed_thickness = $self->surface($surface_num)->set_thickness_with_position($thickness,$position);
		$$zooed_thickness[$position-1]=$thickness;
		$self->set_surface_zoom_data($surface_num,'Thickness',$zooed_thickness);
		$self->set_number_of_position_from_thickness($surface_num);
		return $self;
	}

	sub print_zoom_params {
		my $self = shift;
		my $num_pos = $self->number_of_position;
		if ($num_pos < 2) {
			return;
		}
		else {
			#print "\nNumber of Zoom Position: $num_pos\n";
			print "List of Zoomed Parameters:\n";
			printf " %24s ", q{};
			foreach (1..$num_pos ){
				printf " %11s ", "Position:$_";
			}
			print "\n";
			my $z_sur = $self->{'Zoom'}{'Surfaces'};
			my @keys = keys %$z_sur;
			@keys = sort { $a <=> $b } @keys;
			foreach (@keys){
				printf " Surface %3d",$_;
				my $z_kind = $$z_sur{$_};
				my @keys = keys %$z_kind;
				foreach (@keys) {
					printf "%13s ",$_;
					my $zeemed_params = $$z_kind{$_};
					foreach my $value (@$zeemed_params) {
						if ($value > 1e8) {
							printf " %11g ", $value;
						}
						else {
							printf " %11.4f ", $value;
						}
					}
				}
				print "\n";
			}
			print "\n";
		}
	}

	sub set_etc {
		my $self = shift;
		my $etc = $self->{Etc};
		push(@$etc, @_);
	}

	sub set_indeces {
		my $self = shift;
		$self->reset_position;
		my $glass_catalog = $self->glasscatalog;
		my $w = $self->wavelength;
		if (@_ && $_[0]){
			if ((ref $_[0]) eq 'Wavelength'){
				$w = $glass_catalog->set_indeces($_[0]);
			}
			else {
				$w = $glass_catalog->set_indeces(@_);
			}
		}
		else {
			$w= $glass_catalog->set_indeces($w);
		}
		my $ref_w = $self->reference_wavelength->copy_by_order(1);
		$w = $ref_w->concat($w) unless ($w->get_order($ref_w));
		$self->wavelength($w);
		$self->reset_glasscatalog;
		$self->glasscatalog->set_indeces($w);
		return $self;
	}

	sub reference_wavelength{
		my $self = shift;
		my $reference_wavelength;
		if (@_) {
			$self->{ReferenceWavelength} = shift;
		}
		unless (exists $self->{ReferenceWavelength}) {
			$reference_wavelength = Wavelength->new->by_names('d');
			$self->{ReferenceWavelength} = $reference_wavelength;
		}
		else {
			$reference_wavelength = $self->{ReferenceWavelength};
		}
		if ((ref $reference_wavelength) ne 'Wavelength') {
			$reference_wavelength = Wavelength->new->by_names('d');
			$self->{ReferenceWavelength} = $reference_wavelength;
		}
		return $reference_wavelength;
	}

	sub wavelength {
		my $self = shift;
		if (@_) {
			my $test = shift;
			if ((ref $test) eq 'Wavelength'){
				$self->{Wavelength} = $test;
			}
		}
		$self->{Wavelength} = Wavelength->new->by_names('d') if ((ref $self->{Wavelength}) ne 'Wavelength');
		unless ($self->{Wavelength}->bundle) {
			print STDERR "\n wavelength ERROR\n";
			$self->{Wavelength} = Wavelength->new->by_names('d');
		}
		return $self->{Wavelength};
	}

	sub set_wavelength_by_name {
		my $self = shift;
		my $wavelength;
		if (@_) {
			foreach (@_) {
				push(@$wavelength,Wavelength->wavelength_names($_));
			}
		}
		$self->wavelength($wavelength);
		$self->set_indeces;
		return $self;
	}
	
	sub set_wavelength_by_value {
		my $self = shift;
		my $wavelength;
		$wavelength = Wavelength->new->by_wavelengths($_) if (@_);
		$self->wavelength($wavelength);
		$self->set_indeces;
		return $self;
	}

	sub wavelength_names {
		my $self = shift;
		Wavelength->wavelength_names(@_);
	}

	sub glasscatalog {
		my $self = shift;
		my $surfaces = $self->surfaces;
		my $wavelength = $self->wavelength;
		print STDERR "\n",(ref $self),"::glasscatalog ERROR about Wavelength." if ((ref $wavelength) ne 'Wavelength');
		my $glass_catalog;
		if (@_) {
			$glass_catalog = shift;
		}
		elsif (exists $self->{GlassCatalog}){
			$glass_catalog = $self->{GlassCatalog};
		}
		else {
			$glass_catalog = GlassCatalog->new;	
		}
		$glass_catalog->wavelength($wavelength);
		$glass_catalog->set_indeces($wavelength);
		foreach my $s (@$surfaces) {
			$glass_catalog->add_glass($s->glass);
		}
		$self->{GlassCatalog} = $glass_catalog;
		return $glass_catalog;
	}

	sub reset_glasscatalog {
		my $self = shift;
		$self->{GlassCatalog} = GlassCatalog->new;
		$self->glasscatalog;
	}

	sub reset_aspherical_type_as_matsui {
		my $self = shift;
		my $sur = $self->surfaces;
		foreach my $n (0..$#{$sur}) {
			if ((ref $$sur[$n]) eq 'AsphericalSurface'){
				$self->replace_surface($n,$self->surface($n)->change_as_matsui);
				#print "\nChange Asphere Type to Matsui.\n",$$sur[$n]->list;
			}
		}
		print "\n\n";
		return $self;
	}

	sub svg_plot {
		my $self = shift;
		my $filename = 'plot.svg';
		$filename = $self->save_as_svg($filename,@_);
		system "open $filename";
	}

	sub save_as_svg {
		my $self = shift;
		my $filename = q{};
		$filename = shift;
		$filename = $self->file_name_to_save($filename, ".svg");
		my $svg = $self->svg(@_);
		my $out = $svg->xmlify;
		open SVGFILE, ">$filename"
			or return "ERROR $!";
		print SVGFILE $out;
		close SVGFILE;
		return $filename;
	}

	sub svg {
		my $self = shift;
		my $svg = shift;
		$svg->comment($self->name);
		my $surs = $self->surfaces;
		my $g = $self->object_surface->glass;
		my $a = $self->entrance_pupil_diameter / 3;
		$a = 1 unless($a);
		my $a0 = $a;
		my $init_x = 100;
		$init_x = shift if (@_);
		my $init_y = 100;
		$init_y = shift if (@_);
		my $ax = $init_x;
		my $th0 = 0;
		my $sag0 = 0;
		my $r1 = q{};
		my $d = q{};
		my $glass_color = "rgb(95, 158, 160)";
		my $stroke = "gray";
		$stroke = shift if (@_);
		my $stroke_width = 0.3;
		$stroke_width = shift if (@_);
		#$stroke_width = 2;
		my $opacity = 0.7;
		my $current_surface_tmp = $self->current_surface_number;
		
		foreach my $s (@$surs) {
			next if ($s->number < 1);
			my $th = $s->thickness;
			$a = $s->semidiameter if ($s->semidiameter);
			my $sag = 0;
			if ($self->surface($s->number + 1)->semidiameter && $a < $self->surface($s->number + 1)->semidiameter) {
				$a = $self->surface($s->number + 1)->semidiameter;
			}
			$ax = $ax + $th0;
			my $not_stop = 1;
			if ($s->stop) {
				if ($s->semidiameter) {
					$not_stop = 0;
					my $stop_width = $a * 0.3;
					my $stop_stroke_width = $stroke_width *2;
					$svg->path(d=>" M $ax $init_y m0,-$a l0,-$stop_width ",style=>{stroke=>$stroke,fill=>"none",'stroke-width'=>$stop_stroke_width, 'opacity'=>$opacity});
					$svg->path(d=>" M $ax $init_y m0, $a l0, $stop_width ",style=>{stroke=>$stroke,fill=>"none",'stroke-width'=>$stop_stroke_width, 'opacity'=>$opacity});
				}
				else {
					$not_stop = 1;
				}
			}
			if ((ref $g) eq 'Air' && (ref $s->glass) eq 'Air' && $s->reflect){
				# reflect surface
				$d = "M $ax $init_y ".$s->svg_path('R1',$a)->{SVGPath};
				my $ref_stroke_width = $stroke_width * 3;
				$svg->path(d=>$d,style=>{stroke=>"gray",fill=>"none", 'stroke-width'=>$ref_stroke_width, 'opacity'=>$opacity});
				$sag = 0;
			}
			elsif ((ref $g) eq 'Air' && (ref $s->glass) eq 'Air' && $not_stop){
				# dummy surface
				$svg->path(d=>" M $ax $init_y m0,-$a l0,$a 0,$a ",style=>{stroke=>$stroke,fill=>"none",'stroke-width'=>$stroke_width, 'opacity'=>$opacity});
				$sag = 0;
			}
			elsif ((ref $g) ne 'Air') {
				# R2 surface
				my $sur_svg_path = $s->svg_path('R2',$a0);
				$sag = $sur_svg_path->{Sag};
				my $eg = $th0 - $sag0 + $sag;
				#print STDERR "\n",$s->number," eg:$eg th0:$th0 sag0:$sag0 sag:$sag a0:$a0";
				$d = "$r1 l $eg,0 ".$sur_svg_path->{SVGPath}." z ";
				$svg->path(d=>$d,style=>{stroke=>"rgb(200,200,200)",fill=>$glass_color,'stroke-width'=>$stroke_width, 'opacity'=>$opacity});
				$r1 = q{};
				$sag = 0;
			}
			if ((ref $s->glass) ne 'Air') {
				# $1 Surface
				my $sur_svg_path = $s->svg_path('R1',$a);
				$r1 = "M $ax $init_y ".$sur_svg_path->{SVGPath};
				#print STDERR "\n",$s->number," sag:$sag edge_sag:",$sur_svg_path->{Sag};
				$sag = $sur_svg_path->{Sag};
				$glass_color = $s->glass->svg_fill;
			}
			$g = $s->glass;
			$th0 = $th;
			$sag0 = $sag;
			$a0 = $a;
		}
	#	$svg->path(id=>"axix",  d=>"M 0,$init_y h $ax", style=>{stroke=>'black',fill=>"none",'stroke-width'=>$stroke_width});
		return $svg;
	}

	sub optical_axis {
		my $self = shift;
		$self->{OpticalAxis} = shift if (@_);
		$self->{OpticalAxis};
	}

	sub name {
		my $self = shift;
		if (@_) {
			my $value = shift;
			$self->set_name($value);
		}
		$self->{Name};
	}

	sub set_name {
		my $self = shift;
		my $value = shift;
		$value =~ s/^\s+//;
		$value =~ s/\s+$//;
		$self->{Name} = $value;
		return $self;
	}

	sub safe_name {
		my $self = shift;
		my $name = $self->{Name};
		my $safe_name = q{};
		$name =~ s/[\s\-\.\:\+\;\/\\]+/\_/g;
		while ($name =~ s/([A-Za-z0-9\_]+)// ){
			$safe_name = $safe_name.$1;
		}
		$name =~ s/^\_+//;
		$name =~ s/\_+$//;
		$safe_name = "untitled_lens" if ($safe_name eq q{});
		return $safe_name;
	}

	sub file_name_to_save {
		my $self = shift;
		my $filename = shift;
		if ($filename) {
			$filename =~ s/\s//gi;
			$filename =~ s/\:/_/gi;
			$filename =~ s/\\/_/gi;
			$filename =~ s/\,/_/gi;
		}
		$filename = $self->safe_name unless ($filename);
		$filename = "dumped_lensdata" unless ($filename);
		print "filename:$filename\n";
		my $suf = q{};
		$suf = shift if (@_);
		$suf = ".$suf" if ($suf && $suf !~ m/^\./ );
		unless ($filename =~ m/$suf$/i ) {
			$filename = $filename."$suf" if ($suf);
		}
		return $filename;
	}

	sub file_name_to_save_ {
		my $self = shift;
		my $filename = shift;
		if ($filename) {
			$filename =~ s/\s//gi;
			$filename =~ s/\:/_/gi;
			$filename =~ s/\\/_/gi;
			$filename =~ s/\,/_/gi;
		}
		$filename = $self->safe_name unless ($filename);

		$filename = "dumped_lensdata" unless ($filename);
		my $suffix = q{};
		$suffix = shift if (@_);
		$suffix = ".$suffix" if ($suffix && $suffix !~ m/^\./ );
		print "filename:$filename\n";
		if ($filename =~ m/(.*)(\..+$)/) {
			my $suf = $2;
			my $fname = $1;
			$filename = $fname if ($suffix eq $suf);
		}
		$filename = $filename.$suffix;
		return $filename;
	}

	sub dump {
		my $self = shift;
		use Data::Dumper;
		my $dumper_data = Data::Dumper->Dump([$self],["LENS_DATA"]);
		print $dumper_data;
		return $self;
	}

	sub save_dump {
		my $self = shift;
		my $filename = q{};
		$filename = shift if (@_);
		my $glass_catalog = GlassCatalog->new;
		my $result = $self->save_dump_with_glasscatalog($filename,$glass_catalog);
		return $result;
	}

	sub save_dump_with_glasscatalog {
		my $self = shift;
		my $filename = q{};
		$filename = shift;
		$filename = $self->file_name_to_save($filename);
		my $glass_catalog = shift;
		my $surfaces = $self->surfaces;
		my $number_of_surfaces = @$surfaces;
		if ($number_of_surfaces < 1) {
			return "ERROR No Data. Do not save.";
		}
		foreach my $s (@$surfaces) {
			$glass_catalog->add_glass($s->glass);
		}
		my $zoom_data = $self->zoom;
		$self->reset_position;
		use Data::Dumper;
		my $dumper_data = Data::Dumper->Dump(
			[$glass_catalog,$zoom_data,$self],
			[
				"GLASS_CATALOG_DATA",
				"ZOOM_DATA",
				"LENS_DATA",
			]
			);	
		open DUMPFILE, ">$filename"
			or return "ERROR $!";
		print DUMPFILE $dumper_data;
		close DUMPFILE;
		return $filename;
	}

	sub save_as_yaml {
		my $self = shift;
		my $filename = q{};
		$filename = shift;
		$filename = $self->file_name_to_save($filename, ".yaml");
		my $surfaces = $self->surfaces;
		my $number_of_surfaces = @$surfaces;
		if ($number_of_surfaces < 1) {
			return "ERROR No Data. Do not save.";
		}
		$self->{Position} = {};

		require YAML;
		YAML::DumpFile($filename,$self);

		return $filename;
	}

	sub save_as_xml {
		my $self = shift;
		my $filename = q{};
		$filename = shift;
		$filename = $self->file_name_to_save($filename, ".xml");
		my $surfaces = $self->surfaces;
		my $number_of_surfaces = @$surfaces;
		if ($number_of_surfaces < 1) {
			return "ERROR No Data. Do not save.";
		}
		$self->{Position} = {};

		require XML::Dumper;;
		my $dump = new XML::Dumper;
		$dump->pl2xml( $self, $filename );

		return $filename;
	}

	sub save_as_json {
		my $self = shift;
		my $filename = q{};
		$filename = shift;
		$filename = $self->file_name_to_save($filename, ".json");
		my $surfaces = $self->surfaces;
		my $number_of_surfaces = @$surfaces;
		if ($number_of_surfaces < 1) {
			return "ERROR No Data. Do not save.";
		}
		$self->{Position} = {};

		require JSON;
		my $json = new JSON;
		#$json->indent;
		$json->pretty;
		open JSONFILE, ">$filename"
			or return "ERROR $!";
		print JSONFILE $json->allow_blessed->convert_blessed->encode( $self );
		close JSONFILE;
		return $filename;
	}

	sub TO_JSON {
		#http://d.hatena.ne.jp/koizuss/20090605/1244174060
		my $self = shift;
		my %h_self = map { $_ => $self->{$_} } keys(%$self);
		# インスタンスをhashに変換

		return { ref($self) => \%h_self };    # パッケージ名を付与して返却
	}

	sub zero {
		my $value = shift;
		$value = shift if (ref $value);
		my $limit = 1e-14;
		#$limit = 1e-8;
		$value = 0 if ($value < $limit && $value > -$limit);
		return $value;
	}
	1;
}

