{ package LensShell;
	# ONO Hiroki onohiroki@cup.com 
	# http://www.opensource.org/licenses/bsd-license
	# Copyright (c) 2011, ONO Hiroki onohiroki@cup.com 
	# All rights reserved.
	use strict;
	use warnings;
	require Term::ShellUI;
	use Data::Dumper;
	sub new {
		my $class = shift;
		my $glass_catalog_array = [];
		$glass_catalog_array = shift if (@_);
		my $self = {
			ImportedGlassCatalog => $glass_catalog_array,
			Glass => Air->get,
			GlassCatalog => GlassCatalog->new,
			Surface => Surface->new(0,0,0),
			Lens => Lens->new,
			RaySegment => RaySegment->new,
			Ray => Ray->new,
			Wavelength => Wavelength->new,
			Term => undef,
		};				
		bless $self, $class;
		my $term = new Term::ShellUI(
				prompt => 'PerlLens> ',
				commands => {
						"cd" => {
							desc => "Change to directory DIR",
							maxargs => 1, args => sub { shift->complete_onlydirs(@_); },
							proc => sub { chdir($_[0] || $ENV{HOME} || $ENV{LOGDIR}); },
						},
						"chdir" => { alias => 'cd' },
						"ls" => {
							desc => "List directory",
							maxargs => 0, proc => sub { system('ls -F'); },
						},
						"dir" => { alias => 'ls' },
						"pwd" => {
							desc => "Print the current working directory",
							maxargs => 0, proc => sub { system('pwd'); },
						},
						"run" => {
							desc => "Eval Perl script.",
							maxargs => 1, 
							minargs => 1,
							proc => sub {
								my $file = shift;
								open(IN, "<$file") or die "Can't open file: $!";
								my @input = <IN>;
								close(IN);
								my $data = "@input";
								eval($data);
							},
						},
						"quit" => {
							desc => "Quit this program", maxargs => 0,
							method => sub { shift->exit_requested(1); },
						},
						"exit" => { alias => 'quit' },
						"read" => {
							desc => "Read lens data file",
							maxargs => 1, 
							minargs => 1, 
							proc => sub {
								$glass_catalog_array = $self->imported_glass_catalog;
								my $lens = Lens->read_file($_[0],@$glass_catalog_array);
								$self->lens($lens);
								$self->surface($lens->last_surface);
								$self->wavelength($lens->wavelength);
								$self->glasscatalog($lens->glasscatalog($self->glasscatalog));
							},
						},
						"res" => { alias => 'read' },
						"save" => {
							desc => "Save lens data file",
							maxargs => 3, 
							minargs => 1, 
							proc => sub {
								my $lens = $self->lens;
								my $glass_catalog = $self->glasscatalog;
								my $file = shift;
								my $type = q{};
								my $outputfile = q{};
								if (@_) {
									$type = $file;
									$file = shift;
									if ($type =~ m/^AS$/i) {
										$type = $file;
										$file = shift;
									}
								}
								if ($type =~ m/XML/i) {
									$outputfile = $lens->save_as_xml($file);
								}
								elsif ($type =~ m/YAML/i) {
									$outputfile = $lens->save_as_yaml($file);
								}
								elsif ($type =~ m/JSON/i) {
									$outputfile = $lens->save_as_json($file);
								}
								elsif ($type =~ m/glass/i) {
									$outputfile = $lens->save_dump_with_glasscatalog($file,$glass_catalog);
								}
								else {
									$outputfile = $lens->save_dump($file);
								}
								print "Save as $outputfile\n";
							},
						},
						"set" => {
							desc => "Set index or aspherical surface or ...",
							proc => sub {
								my $lens = $self->lens;
								my $glass_catalog = $self->glasscatalog;
								my $wavelength = $self->wavelength;
								my $type = shift;
								if ($type eq 'indeces' || $type eq 'index' || $type eq 'wavelength') {
									$lens->reset_position;
									my @names = @_;
									@names = ('C','d','F') if ($#names < 1);
									$wavelength = Wavelength->new->by_wavelengths(@_)->list;
									$lens->wavelength($wavelength);
									$lens->set_indeces;
								}
								elsif ($type =~ m/^asp/i ) {
									$lens->reset_aspherical_type_as_matsui;
								}
								else {
									print "Error: Not implement:$type\n";
								}
							},
						},
						"reset" => {
							desc => "Reset index or aspherical type or diameter or ...",
							proc => sub {
								my $lens = $self->lens;
								my $wavelength = $self->wavelength;
								my $ray = $self->ray;
								my $type = shift;
								if ($type eq 'indeces' || $type eq 'index' || $type eq 'wavelength') {
									$lens->reset_position;
									my @names = @_;
									@names = ('C','d','F') if ($#names < 1);
									$wavelength = Wavelength->new->by_wavelengths(@_)->list;
									$lens->wavelength($wavelength);
									$lens->set_indeces;
								}
								elsif ($type =~ m/^asp/i ) {
									$lens->reset_aspherical_type_as_matsui;
								}
								elsif ($type =~ m/^dia/i ) {
									
									my $h = ($lens->entrance_pupil_diameter)/2;
									my @rays = ();
									my $f1r1 = RaySegment->new([1,0,0],[0,0,0],  [1,0,0],0);
									my $f1r2 = RaySegment->new([1,0,0],[0,$h,0], [1,0,0],0);
									foreach ($f1r1, $f1r2) {
										$ray = $lens->paraxial_ray_trace($_);
										push (@rays, $ray);
									}
									foreach my $r (@rays) {
										my $segments = $r->segments;
										foreach my $s (@$segments) {
											my $num = $s->num_of_surface;
											my $t = $s->get_t;
											my $map = abs(${$s->get_t}[1]);
											$lens->surface($num)->semidiameter($map + 1);
										}
									}
								}
								else {
									print "Error: Not implement:$type\n";
								}
							},
						},
						"sav" => { alias => 'save' },
						"new"=> {
							desc => "Create new lens.",
							maxargs => 0, 
							proc => sub {
								$self->lens(Lens->new);
							},
						},
						"glass"=>{
							desc => "Create new glass.",
							proc => sub {
								my $glass = $self->glass;
								my $glass_catalog = $self->glasscatalog;
								my $num_of_args = @_;
								if ($num_of_args == 4){
									$glass = $glass_catalog->add_new_glass(@_);
								}
								else {
									$glass_catalog->list;
									print "\nCurrent Glass:";
									$glass->list;
								}
							},
						},
						"g" => { alias => 'glass' },
						"glasscatalog"=>{
							desc => "Show Glass Catalog.",
							maxargs => 1, 
							proc => sub {
								my $lens = $self->lens;
								my $glass_catalog = $self->glasscatalog;
								if (@_) {
									my $test = shift;
									if ($test =~ m/new/i){
										$glass_catalog = GlassCatalog->new;
									}
									elsif ($test =~ m/dump/i) {
										print Dumper($glass_catalog);
									}
									elsif ($test =~ m/test/i){
										$lens->glasscatalog->test_index_dispersion_formula;
									}
									elsif ($test eq 'file') {
										my $type = shift;
										if ($type =~ m/^afg$/i) {
											my $agf_glass_catalog = GlassAGF->read;
											$agf_glass_catalog->save_dump("AGFGLASS");
											$glass_catalog->concat($agf_glass_catalog);
										}
										if ($type =~ m/^dat$/i) {
											my $agf_glass_catalog = CODEV->read_glass_macro_data;
											$agf_glass_catalog->save_dump("CVGLASS");
											$glass_catalog->concat($agf_glass_catalog);
										}
									}
									print "\n";
								}
								else {
									$glass_catalog->list;
								}
							},
						},
						"surface"=>{
							desc => "Create new surface.",
							proc => sub {
								my $lens = $self->lens;
								my $surface = $self->surface;
								my $glass = $self->glass;
								my $glass_catalog = $self->glasscatalog;
								my $surfaces;
								my $num_of_args = @_;
								if ($num_of_args == 1) {
									my $num = shift;
									if ($num =~ m/^last$/i || $num =~ m/^im/i){
										$num = $lens->last_surface->number;
									}
									elsif ($num =~ m/^first$/ || $num =~ m/^obj/i){
										$num = $lens->first_surface->number;
									}
									elsif ($num =~ m/^stop?$/){
										$num = $lens->stop_surface->number;
									}
									$lens->current_surface($lens->surface($num));
									$surface = $lens->current_surface;
									print "Current Surface:",$surface->list, "\n";
								}
								elsif (@_) {
									my $r = shift;
									my $t = shift;
									my $g;
									if (@_) {
										$g = shift;
										$glass = $glass_catalog->glass($g);
										$surfaces = $lens->add_surface($r, $t, $glass);
									}
									else {
										$surfaces = $lens->add_surface($r, $t);
									}
									print "Surface:",$lens->current_surface->number,"\n";
								}
								else {
									$surface = Surface->new(0,0,0) unless ($surface);
									print "Current Surface:",$surface->list, "\n";
								}
							},
						},
						"s" => { alias => 'surface' },
						"delete" => {
							desc => "Delete Data.",
							maxargs => 2, 
							proc => sub {
								my $surface = $self->surface;
								my $lens = $self->lens;
								my $type = shift;
								if ($type =~ m/^Surface$/i || $type =~ m/^S$/i ){
									my $num_sur = shift;
									$surface = $lens->delete_surface($num_sur);
								}
								else {
									print "Error: Not implement.\n";
								}
							},
						},
						"list"=> {
							desc => "Show Lens data list.",
							maxargs => 1, 
							proc => sub {
								my $glass_catalog = $self->glasscatalog;
								my $lens = $self->lens;
								my $ray = $self->ray;
								my $wavelength = $self->wavelength;
								if (@_) {
									my $type = shift;
									if ($type eq 'glass') {
										$glass_catalog->list;
									}
									elsif ($type =~ m/^sur/) {
										$lens->list_surfaces;
									}
									elsif ($type eq 'zoom') {
										$lens->print_zoom_params;
									}
									elsif ($type eq 'ray') {
										$ray->list;
									}
									elsif ($type eq 'wavelength'){
										if ((ref $lens->reference_wavelength) eq 'Wavelength') {
											print "Lens Refrence:";
											$lens->reference_wavelength->list;
										}
										if ((ref $lens->wavelength) eq 'Wavelength') {
											print "Lens:";
											$lens->wavelength->list;
										}
										print "\nShell Local:";
										$wavelength->list;
										print "\n";
									}
									else {
										print "Error: Not implement:$type\n";
									}
								}
								else {
									$lens->set_indeces($wavelength);
									$lens->list;
								}
							},
						},
						"lis" => { alias => 'list' },
						"position"=> {
							desc => "Change lens zoom position.",
							maxargs => 1, 
							minargs => 1, 
							proc => sub {
								my $pos = shift;
								my $lens = $self->lens;
								$lens->current_position($pos);
							},
						},
						"pos" => { alias => 'position' },
						"paraxial_ray_trace"=> {
							desc => "Paraxial Ray Trace",
							maxargs => 4, 
							minargs => 3, 
							proc => sub {
								my $lens = $self->lens;
								my $ray = $self->ray;
								my $ray_segment = $self->ray_segment;
								my $glass = $self->glass;								
								$ray_segment = RaySegment->new([1,0,0],[0,$_[0],0],[1
									,$_[1],0],$_[2]);
								$ray_segment->wavelength($glass->wavelength_names($_[3])) if ($_[3]);
								$ray = $lens->paraxial_ray_trace($ray_segment)->list;
							},
						},
						"pray" => { alias => 'paraxial_ray_trace' },
						"skew_ray_trace"=> {
							desc => "Skew Ray Trace",
							maxargs => 8, 
							minargs => 6, 
							proc => sub {
								my $surface = $_[6];
								$surface = 0 unless ($surface);
								my $wavelength_name = $_[7];
								my $ray_segment = $self->ray_segment;
								my $glass = $self->glass;
								my $ray = $self->ray;
								my $lens = $self->lens;
								$ray_segment = RaySegment->new([1,0,0],[$_[0],$_[1],$_[2]],[$_[3],$_[4],$_[5]],$surface);
								$ray_segment->wavelength($glass->wavelength_names($wavelength_name)) if ($wavelength_name);
								$ray = $lens->skew_ray_trace($ray_segment)->list;
							},
						},
						"sray" => { alias => 'skew_ray_trace' },
						"plot" => {
							desc => "Draw lens.",
							proc => sub {
								require Lens::PlotSVG;
								my $lens = $self->lens;
								$lens->paraxial;
								my $deg = 0;
								my $v1 = 1;
								my $v2 = 1;
								my $plot_scale = 3;
								$plot_scale = shift if (@_);
								$deg = shift if (@_);
								$v1 = shift if (@_);
								$v2 = shift if (@_);
								my $vig = $lens->vignetting;
								my $h = ($lens->entrance_pupil_diameter)/2;
								my $svg = PlotSVG->new(init_x=>10,init_y=>40,lens_scale=>$plot_scale);
								my @rays = ();
								my $f1r1 = RaySegment->new([1,0,0],[0,0,0],  [1,0,0],0);
								my $f1r2 = RaySegment->new([1,0,0],[0,$h,0], [1,0,0],0);
								my $f1r3 = RaySegment->new([1,0,0],[0,-$h,0], [1,0,0],0);
								my $ray = $self->ray;
								foreach ($f1r1, $f1r2, $f1r3) {
									$ray = $lens->skew_ray_trace($_);
									$svg->add($ray,'red');
									push (@rays, $ray);
								}
								if ($deg) {
									my $pi = atan2(1,1) * 4;
									my $deg2rad = 1 / 360 * 2 * $pi;
									my $tan = sin($deg * $deg2rad) / cos($deg * $deg2rad);
									my $x = 1 / (1 + $tan**2)**0.5;
									my $y = $x * $tan;
									my $f2r1 = RaySegment->new([1,0,0],[0,0,0],  [$x,$y,0],-1);
									my $f2r2 = RaySegment->new([1,0,0],[0,$h*$v1,0], [$x,$y,0],-1);
									my $f2r3 = RaySegment->new([1,0,0],[0,-$h*$v2,0],[$x,$y,0],-1);
									foreach ($f2r1, $f2r2,$f2r3) {
										$ray = $lens->skew_ray_trace_with_ep($_);
										$svg->add($ray,'green');
										push (@rays, $ray);
									}
								}
								foreach my $r (@rays) {
									my $segments = $r->segments;
									foreach my $s (@$segments) {
										my $num = $s->num_of_surface;
										my $t = $s->get_t;
										my $map = abs(${$s->get_t}[1]);
										$lens->surface($num)->semidiameter($map + 1) if ($lens->surface($num)->semidiameter < $map);
									}
								}
								$svg->add($lens);
								$svg->plot;
							}
						},
						"dump"=> {
							desc => "Show lens data as dump.",
							maxargs => 1, 
							proc => sub {
								my $lens = $self->lens;
								my $glass = $self->glass;
								my $glass_catalog = $self->glasscatalog;
								my $ray = $self->ray;
								my $ray_segment = $self->ray_segment;
								my $wavelength = $self->wavelength;
								my $surface = $self->surface;
								$lens->dump unless (@_);
								if (@_){
									my $type = shift;
									if ($type eq 'lens') {
										$lens->dump;
									}
									elsif ($type eq 'glass'){
										$glass->dump;
									}
									elsif ($type eq 'glasscatalog'){
										$glass_catalog->dump;
									}
									elsif ($type eq 'ray'){
										$ray->dump;
									}
									elsif ($type eq 'raysegment'){
										$ray_segment->dump;
									}
									elsif ($type eq 'surface'){
										$surface->dump;
									}
									elsif ($type eq 'wavelength'){
										$wavelength->dump;
									}
									else {
										print "Error: Not implement:$type\n";
									}
								}
							},
						},
						"yaml"=> {
							desc => "Show lens data as YAML.",
							maxargs => 0, 
							proc => sub {
								my $lens = $self->lens;
								require YAML;
								print YAML::Dump($lens);
							},
						},
						"json"=> {
							desc => "Show lens data as JSON.",
							maxargs => 0, 
							proc => sub {
								my $lens = $self->lens;
								require JSON;
								my $json = new JSON;
								$json->pretty;
								print $json->allow_blessed->convert_blessed->encode( $lens );
							},
						},
						"?" =>		{ syn => "help" },
						"help" => 	{ 
							desc => "Print helpful information",
							args => sub { shift->help_args(undef, @_); },
							meth => sub { shift->help_call(undef, @_); }
						},			
						"about"=> {
							desc => "Show License",
							maxargs => 0, 
							proc => sub {
								my $lens = $self->lens;
								$lens->license;
							},
						},
						"license" => { alias => 'about' },
					},
				history_file => '~/.lens-shellui-synopsis-history',
			);
		$self->term($term);
		return $self;
	}

	sub term {
		my $self = shift;
		$self->{Term} = shift if (@_);
		return $self->{Term};
	}

	sub surface {
		my $self = shift;
		$self->{Surface} = shift if (@_);
		return $self->{Surface};
	}

	sub ray_segment {
		my $self = shift;
		$self->{RaySegment} = shift if (@_);
		return $self->{RaySegment};
	}

	sub wavelength {
		my $self = shift;
		$self->{Wavelength} = shift if (@_);
		return $self->{Wavelength};
	}

	sub ray {
		my $self = shift;
		$self->{Ray} = shift if (@_);
		return $self->{Ray};
	}

	sub lens {
		my $self = shift;
		$self->{Lens} = shift if (@_);
		return $self->{Lens};
	}

	sub glass {
		my $self = shift;
		$self->{Glass} = shift if (@_);
		return $self->{Glass};
	}

	sub glasscatalog {
		my $self = shift;
		$self->{GlassCatalog} = shift if (@_);
		return $self->{GlassCatalog};
	}

	sub imported_glass_catalog {
		my $self = shift;
		$self->{ImportedGlassCatalog} = shift if (@_);
		return $self->{ImportedGlassCatalog};
	}

	sub print_imported_glass_catalog_number {
		my $self = shift;
		my $glass_catalog_array = $self->{ImportedGlassCatalog};
		my $num = 0;
		map { $num++ if ((ref $_) eq 'GlassCatalog') } @$glass_catalog_array;
		print "$num Glass Catalog Imported.\n" if ($num);
		return $self;
	}

	sub run {
		my $self = shift;
		print 'Using '.$self->term->{term}->ReadLine."\n";
		print "Welcome to Perl Lens Shell.\n";
		$self->print_imported_glass_catalog_number;
		$self->term->run(@_);
	}
	1;
}

	  