{ package PlotSVG;
	# ONO Hiroki onohiroki@cup.com 
	# http://www.opensource.org/licenses/bsd-license
	# Copyright (c) 2011, ONO Hiroki onohiroki@cup.com 
	# All rights reserved.
	use strict;
	use warnings;
	use SVG;
	@PlotSVG::ISA = qw(SVG);
	sub new {
		my $class = shift;
		my $init_x = 100;
		my $init_y = 100;
		my $lens_plot_scale = 3;
		#$init_x = shift if (@_);
		#$init_y = shift if (@_);
		my @a = ();
		while (@_) {
			my $test = shift;
			if ($test eq 'init_x') {
				$init_x = shift;
			}
			elsif ($test eq 'init_y') {
				$init_y = shift;
			}
			elsif ($test eq 'lens_scale') {
				$lens_plot_scale = shift;
			}
			else {
				push(@a,$test);
			}
		}
		print "\nPlotSVG::new init_x:$init_x init_y:$init_y\n@a\n";
		my $svg = SVG->new (@a);
		my $lens_plot = $svg->group(id=>'lens_plot',transform=>"scale($lens_plot_scale)");
		my @keys = keys %$svg;
		my $self;
		foreach (@keys) {
			$$self{$_} = $$svg{$_};
		}
		$$self{init_x} = $init_x;
		$$self{init_y} = $init_y;
		bless $self, $class;
		$self->{LensPlot} = $lens_plot;
		$self->{LensPlotScale} = $lens_plot_scale;
		return $self;
	}

	sub lens_plot {
		my $self = shift;
		$self->{LensPlot};
	}

	sub lens_plot_scale {
		my $self = shift;
		$self->{LensPlotScale};
	}

	sub init_x {
		my $self = shift;
		$$self{init_x} = shift if (@_);
		return $$self{init_x};
	}

	sub init_y {
		my $self = shift;
		$$self{init_y} = shift if (@_);
		return $$self{init_y};
	}

	sub add {
		my $self = shift;
		my $obj = shift;
		if ((ref $obj) =~ m/Lens/i ) {
			$self->add_lens($obj, @_);
		}
		elsif ((ref $obj) eq "Ray" ) {
			$self->add_ray($obj, @_);
		}
		else {
			if (ref $obj) {
				print STDERR "\nERROR ", (ref $self), "::add_ray_path ",(ref $obj), " is not Lens Object.";
			}
			else {
				print STDERR "\nERROR ", (ref $self), "::add_ray_path $obj is not Lens Object.";
			}
		}
		return $self;
	}

	sub add_lens {
		my $self = shift;
		my $lens = shift;
		my $group = $self->lens_plot;
		my $stroke = "gray";
		$stroke = shift if (@_);
		my $stroke_width = 0.5;
		$stroke_width = $stroke_width * (3 / $self->lens_plot_scale);
		$stroke_width = shift if (@_);
		my $init_x = $self->init_x;
		my $init_y = $self->init_y;
		$init_y = shift if (@_);
		if ((ref $lens) !~ m/Lens/i) {
			print STDERR "\nERROR ", (ref $self), "::add_ray_path $lens is not Lens Object.";
			return $self;
		}
		$self = $lens->svg($group,$init_x,$init_y,$stroke,$stroke_width);
		return $self;
	}

	sub add_ray {
		my $self = shift;
		my $ray = shift;
		my $group = $self->lens_plot;
		my $stroke = "gray";
		$stroke = shift if (@_);
		my $stroke_width = 0.5;
		$stroke_width = $stroke_width * (3 / $self->lens_plot_scale);
		$stroke_width = shift if (@_);
		if ((ref $ray) ne "Ray") {
			print STDERR "\nERROR ", (ref $self), "::add_ray_path $ray is not Ray Object.";
			return $self;
		}
		my $init_x = $self->init_x;
		my $init_y = $self->init_y;
		my $svg_ray_path = $ray->svg_path($init_x,$init_y);
		$group->path(d=>$svg_ray_path,style=>{stroke=>$stroke,fill=>"none",'stroke-width'=>$stroke_width});
		return $self;
	}

	sub plot {
		my $self = shift;
		my $something = q{};
		if (@_) {
			$something = shift;
			$something =~ s/\s+/_/;
			$something =~ s/[^A-Za-z0-9_]//g;
		}
		my $filename = 'plot'.$something.'.svg';
		$filename = $self->save_as_file($filename,@_);
		if (-x "/usr/bin/open") {
			if (-x "/Applications/Safari.app") {
				system "/usr/bin/open","-a","/Applications/Safari.app",$filename;
			}
			else {
				system "/usr/bin/open","$filename";
			}
		}
		return $self;
	}

	sub save_as_file {
		my $self = shift;
		my $filename = q{};
		$filename = shift;
		$filename = $self->Lens::file_name_to_save($filename, ".svg");
		my $out = $self->xmlify;
		open SVGFILE, ">$filename"
			or return "ERROR $!";
		print SVGFILE $out;
		close SVGFILE;
		return $filename;
	}
	
	1;
}
