{ package RayBundle;
	# ONO Hiroki onohiroki@cup.com 
	# http://www.opensource.org/licenses/bsd-license
	# Copyright (c) 2011, ONO Hiroki onohiroki@cup.com 
	# All rights reserved.
	use strict;
	use warnings;
	sub new {
		my $class = shift;
		my $self = {
			RayBundle => [],
			OpticalAxis => [1,0,0],
		};
		bless $self, $class;
	}

	sub new_fan {
		my $class = shift;
		my $num_of_rays = shift;
		my $lens;
		$lens = shift if (@_);
		my $h = ($lens->entrance_pupil_diameter)/2;
		my $i = $h / $num_of_rays;
		my $s = $h;
		my $field_angle = $lens->field_angle;
		my @angles = map {-($$_[0]**2 + $$_[1]**2)**0.5} @$field_angle;
		my $pi = atan2(1,1) * 4;
		my $deg2rad = 1 / 360 * 2 * $pi;
		my $opticalaxis = [1,0,0];
		my @rays = ();
		foreach my $deg (@angles) {
			my $tan = sin($deg * $deg2rad) / cos($deg * $deg2rad);
			my $a = 1 / (1 + $tan**2)**0.5;
			my $b = $a * $tan;
			while ($s >= -$h) {
				my $ray = RaySegment->new($opticalaxis,[0,$s,0],  [$a,$b,0],0);
				push (@rays, $ray);
				$s = $s - $i;
			}
		}
		my $self = {
			RayBundle => \@rays,
			OpticalAxis => $opticalaxis,
		};
		bless $self, $class;
	}
	
	sub new_y_fan {
		my $class = shift;
		my $num_of_rays = shift;
		my $lens = shift;
		my $h = ($lens->entrance_pupil_diameter)/2;
		my $fields = $lens->field;
		my $pi = atan2(1,1) * 4;
		my $deg2rad = 1 / 360 * 2 * $pi;
		my $opticalaxis = [1,0,0];
		my @rays = ();
		foreach my $f (@$fields) {
			my $deg = $f->{Angle}[0];
			my $vig1 = $f->{Vignetting}[0];
			my $vig2 = $f->{Vignetting}[1];
			my $s = $h * $vig1;
			my $i = ($s + $h * $vig2) / ($num_of_rays -1);
			my $tan = sin($deg * $deg2rad) / cos($deg * $deg2rad);
			my $a = 1 / (1 + $tan**2)**0.5;
			my $b = $a * $tan;
			while ($s >= -$h * $vig2) {
				my $rayseg = RaySegment->new($opticalaxis,[0,$s,0],  [$a,$b,0],-1);
				my $ray = $lens->skew_ray_trace_with_ep($rayseg);
				push (@rays, $ray);
				$s = $s - $i;
			}
		}
		my $self = {
			RayBundle => \@rays,
			OpticalAxis => $opticalaxis,
		};
		bless $self, $class;
	}

	sub list {
		my $self = shift;
		my $rays = $self->{RayBundle};
		map {$_->list} @$rays;
		return $self;
	}

	sub set_semi_diameter {
		my $self = shift;
		my $lens = shift;
		my $rays = $self->{RayBundle};
		foreach my $r (@$rays) {
			my $segments = $r->segments;
			foreach my $s (@$segments) {
				my $num = $s->num_of_surface;
				my $t = $s->get_t;
				my $map = abs(${$s->get_t}[1]);
				$lens->surface($num)->semidiameter($map + 1) if ($lens->surface($num)->semidiameter < $map);
			}
		}
		return $self;
	}

	sub add_rays_svg_to {
		my $self = shift;
		my $svg = shift;
		my $rays = $self->{RayBundle};
		map {$svg->add($_,'red')} @$rays;
		return $self;
	}

	1;
}

{ package RaySegment;
	# ONO Hiroki onohiroki@cup.com 
	# open source project
	use strict;
	use warnings;
	sub new {
		my $class = shift;
		my $optical_axis = shift;
		my $t = shift;
		my $q = shift;
		my $sur = shift;
		my $thickness = 0;
		$thickness = shift if (@_);
		my $self = {
			T => $t,
			Q => $q,
			OpticalAxis => $optical_axis,
			NumOfSurface => $sur,
			Option=>{},
			Wavelength=>undef,
			Thickness => $thickness,
		};
		bless $self, $class;
	}
	sub new_x {
		my $class = shift;
		my $optical_axis = [1,0,0];
		my $q = [1,0,0];
		my $t = [0,0,0];
		my $self = {
			T => $t,
			Q => $q,
			OpticalAxis => $optical_axis,
			NumOfSurface => 0,
			Option=>{},
			Wavelength=>undef,
			Thickness => 0,
		};
		bless $self, $class;
	}
	sub new_z {
		my $class = shift;
		my $optical_axis = [0,0,1];
		my $q = [1,0,0];
		my $t = [0,0,0];
		my $self = {
			T => $t,
			Q => $q,
			OpticalAxis => $optical_axis,
			NumOfSurface => 0,
			Option=>{},
			Wavelength=>undef,
			Thickness => 0,
		};
		bless $self, $class;
	}
	sub optical_axis {
		my $self = shift;
		$self->{OpticalAxis};
	}
	sub num_of_surface {
		my $self = shift;
		$self->{NumOfSurface} = shift if (@_);
		$self->{NumOfSurface};
	}
	sub set_num_of_surface {
		my $self = shift;
		$self->{NumOfSurface} = shift;
		return $self;
	}
	sub get_q {
		my $self = shift;
		$self->{Q};
	}
	sub get_t {
		my $self = shift;
		$self->{T};
	}
	sub set_t {
		my $self = shift;
		$self->{T} = \@_;
		return $self;
	}
	sub set_q {
		my $self = shift;
		$self->{Q} = \@_;
		return $self;
	}
	sub option {
		my $self = shift;
		$self->{Option} = shift if (@_);
		$self->{Option};
	}

	sub list {
		my $self = shift;
		my $option = $self->{Option};
		if (exists $option->{a}) {
			$self->paraxial_ray_list(@_);
		}
		else {
			$self->skew_ray_list(@_);
		}
		#print " ",$self->thickness;
	}

	sub csv {
		my $self = shift;
		my $q = $self->get_q;
		my $t = $self->get_t;
		my $format = "%8d,%15.6f,%15.6f,%15.6f,%15.6f,%15.6f,%15.6f";
		my $string = sprintf $format, $self->num_of_surface,
						$$t[0],$$t[1],$$t[2],
						$$q[0],$$q[1],$$q[2];
		return $string;
	}

	sub skew_ray_list {
		my $self = shift;
		my $header_mode = shift;
		my $header ="\n";
		$header = shift if (@_);
		my $q = $self->get_q;
		my $t = $self->get_t;
		my $optical_axis = $self->optical_axis;
		if ($header_mode && $header_mode eq 'header') {
			if ($$optical_axis[0] == 0 && $$optical_axis[1] == 0 && $$optical_axis[2] == 1) {
				print "\nOptical Axis: Z"
			}
			else {
				print "\nOptical Axis: X"
			}
			use Data::Dumper;
			print "  Wavelength: ", $self->wavelength_name, " ", $self->wavelength_value, "nm" if ($self->wavelength_name);
			printf "$header %8s %15s %15s %15s %15s %15s %15s", 'Surface#', 'X', 'Y', 'Z', 'a', 'b', 'c';
			return 'header';
		}
		my $format = " %8d %15.6f %15.6f %15.6f %15.6f %15.6f %15.6f";
		$format = " %8d %15.7g %15.7g %15.7g %15.6f %15.6f %15.6f" if (abs($$t[0]) > 1e8 ||abs($$t[1]) > 1e8 ||abs($$t[2]) > 1e8);
		if ($$optical_axis[0] == 0 && $$optical_axis[1] == 0 && $$optical_axis[2] == 1) {
			printf $format, $self->num_of_surface,
				$$t[2],$$t[1],$$t[0],
				$$q[2],$$q[1],$$q[0];
		}
		else {
			printf $format, $self->num_of_surface,
				$$t[0],$$t[1],$$t[2],
				$$q[0],$$q[1],$$q[2];
		}
		return $self;
	}
	sub paraxial_ray_list {
		my $self = shift;
		my $header_mode = shift;
		my $header ="\n";
		$header = shift if (@_);
		my $q = $self->get_q;
		my $t = $self->get_t;
		my $a = $self->{Option}->{a};
		my $optical_axis = $self->optical_axis;
		if ($header_mode && $header_mode eq 'header') {
			print "\nParaxial Ray List:";
			print "  Wavelength:", $self->wavelength_name, " ", $self->wavelength_value if ($self->wavelength_name);
			printf "$header %8s %15s %15s", 'Surface#','h','u';
			return 'header';
		}
		my $format = " %8d %15.6f %15.6f";
		$format = " %8d %15.8g %15.8g " if (abs($$t[1]) > 1e8);
		my $sur_num = 0;
		$sur_num = $self->num_of_surface if ($self->num_of_surface);
		printf $format, $self->num_of_surface, $$t[1],$$q[1];
		return $self;
	}
	sub paraxial_ray_a {
		my $self = shift;
		if (@_) {
			$self->{Option}->{a} = shift;
		}
		if (exists $self->{Option}->{a}) {
			return $self->{Option}->{a};
		}
		else {
			return undef;
		}
	}
	sub paraxial_ray_h {
		my $self = shift;
		return ${$self->get_t}[1];
	}
	sub paraxial_ray_u {
		my $self = shift;
		return ${$self->get_q}[1];
	}

	sub wavelength {
		my $self = shift;
		if (@_) {
			if ((ref $_[0]) ne 'Wavelength') {
				print STDERR "\n",(ref $self),"::wavelength ERROR (",$_[0],") is not object.\n";
				return;
			}
			else {
				$self->{Wavelength} = shift;
				return $self->{Wavelength};
			}
		}
		elsif (exists $self->{Wavelength}){
			return $self->{Wavelength};
		}
		else {
			return undef;
		}
	}

	sub wavelength_name {
		my $self = shift;
		my $w = $self->wavelength;
		if ((ref $w) eq 'Wavelength') {
			return $w->name;
		}
		else {
			return q{};
		}
	}

	sub wavelength_value {
		my $self = shift;
		my $w = $self->wavelength;
		if ((ref $w) eq 'Wavelength') {
			return $w->wavelength;
		}
		else {
			return q{};
		}
	}

	sub reverse_beta {
		my $self = shift;
		my $q = $self->get_q;
		$self->{Q} = [ $$q[0], -$$q[1], $$q[2] ];
		return $self;
	}

	sub thickness {
		my $self = shift;
		$self->{Thickness} = shift if(@_);
		$self->{Thickness}
	}

	sub set_thickness {
		my $self = shift;
		$self->thickness(@_);
		return $self;
	}
	
	sub dump {
		my $self = shift;
		use Data::Dumper;
		my $dumper_data = Data::Dumper->Dump([$self],["RAY_SEGMENT"]);
		print $dumper_data;
		return $self;
	}

	1;
}

{ package Ray;
	# ONO Hiroki onohiroki@cup.com 
	# open source project
	use strict;
	use warnings;
	sub new {
		my $class = shift;
		my $optical_axis = [1,0,0];
		my $self = {
			Segments => [],
			OpticalAxis => $optical_axis,
			CurrentSegment => 0,
		};
		bless $self, $class;
	}
	sub new_with_segment {
		my $class = shift;
		my $segments = [];
		my $optical_axis = [1,0,0];
		my $segment = [];
		if (@_) {
			$segment = shift;
			$optical_axis = $segment->optical_axis;
			push (@$segments, $segment);	
		}
		my $self = {
			Segments => $segments,
			OpticalAxis => $optical_axis,
			CurrentSegment => 0,
		};
		bless $self, $class;
	}
	sub current_segment_number {
		my $self = shift;
		if (@_){
			$self->{CurrentSegment} = shift;
		}
		$self->{CurrentSegment}
	}
	sub increase_current_segment_number{
		my $self = shift;
		$self->{CurrentSegment} = $self->{CurrentSegment}+1;
		$self->{CurrentSegment};
	}
	sub decrease_current_segment_number{
		my $self = shift;
		$self->{CurrentSegment} = $self->{CurrentSegment}-1;
		$self->{CurrentSegment};
	}
	sub add_segment {
		my $self = shift;
		my $segment = shift;
		my $segments = $self->segments;
		$self->{OpticalAxis}=$segment->optical_axis;
		push (@$segments, $segment);
		$self->increase_current_segment_number;
		$segments;
	}
	sub segments {
		my $self = shift;
		$self->{Segments}
	}
	sub segment {
		my $self = shift;
		my $segment = shift;
		my $segments = $self->segments;
		$self->current_segment_number($segment);
		return $$segments[$segment];
	}
	sub reverse_beta_segment {
		my $self = shift;
		my $num = shift;
		my $segments = $self->segments;
		my $segment1 = $$segments[$num];
		my $segment2 = $$segments[$num - 1]->reverse_beta;
		my $optical_axis = $segment1->optical_axis;
		my $t = $segment1->get_t;
		my $q = $segment2->get_q;
		my $sur = $segment1->num_of_surface;
		$self->current_segment_number(0);
		return RaySegment->new($optical_axis,$t,$q,$sur);
	}
	sub optical_axis {
		my $self = shift;
		$self->{OpticalAxis};
	}
	sub list {
		my $self = shift;
		my $segments = $self->segments;
		return $self if (@$segments < 1);
		my $header = sprintf " \n %8s", "Segment#";
		$$segments[0]->list('header', $header);
		my $num = 0;
		foreach my $segment (@$segments) {
			printf " \n %8d", $num++;
			$segment->list;
		}
		print" \n";
		return $self;
	}
	sub paraxial_ray_list {
		my $self = shift;
		my $segments = $self->segments;
		my $header = sprintf " \n %8s", "Segment#";
		$$segments[0]->paraxial_ray_list('header', $header);
		my $num = 0;
		foreach my $segment (@$segments) {
			printf " \n %8d", $num++;
			$segment->paraxial_ray_list;
		}
		print" \n";
		return $self;
	}
	sub paraxial {
		my $self = shift;
		my $seg_obj = $self->first;
		my $seg_img = $self->last;
		my $seg_prev_img = $self->previous_segment;
		my $lens_last_seg = $self->segment(($self->number_of_last)-1);
		my $paraxial_image_distance = 0;
		if (zero($lens_last_seg->paraxial_ray_u) != 0) {
			$paraxial_image_distance = - $lens_last_seg->paraxial_ray_h / $lens_last_seg->paraxial_ray_u ;
		}
		my $focal_length = 0;
		my $beta = 0;
		if (zero($seg_img->paraxial_ray_a) != 0 ) {
			$focal_length = $seg_obj->paraxial_ray_h/$seg_img->paraxial_ray_a;
			$beta = $seg_obj->paraxial_ray_a / $seg_img->paraxial_ray_a;
		}
		return {
			FocalLength=>$focal_length,
			Beta=>$beta,
			ParaxialImageDistance=>$paraxial_image_distance,
			NumberOfLastSegment=>$self->number_of_last,
			LastSegment=>$self->last,
			FirstSegment=>$self->first,
			};
	}
	sub first {
		my $self = shift;
		my $segments = $self->segments;
		$self->current_segment_number(0);
		return $$segments[0];
	}
	sub last {
		my $self = shift;
		my $segments = $self->segments;
		$self->current_segment_number($self->number_of_last);
		return $$segments[$#{$segments}];
	}
	sub number_of_last {
		my $self = shift;
		my $segments = $self->segments;
		return $#{$segments};
	}
	sub next_segment_number {
		my $self = shift;
		return $self->current_segment_number -1;
	}
	sub previous_segment_number {
		my $self = shift;
		return $self->current_segment_number -1;
	}
	sub next_segment{
		my $self = shift;
		my $segments = $self->segments;
		return $$segments[$self->next_segment_number];
	}
	sub previous_segment {
		my $self = shift;
		my $segments = $self->segments;
		return $$segments[$self->previous_segment_number];
	}
	sub current_segment {
		my $self = shift;
		my $segments = $self->segments;
		return $$segments[$self->current_segment_number];
	}

	sub get_h_of_segment {
		my $self = shift;
		my $num = $self->number_of_last;
		$num = shift if (@_);
		my $segments = $self->segments;
		return $$segments[$num]->paraxial_ray_h;
	}
	sub get_u_of_segment {
		my $self = shift;
		my $num = $self->number_of_last;
		$num = shift if (@_);
		my $segments = $self->segments;
		return $$segments[$num]->paraxial_ray_u;
	}
	sub get_a_of_segment {
		my $self = shift;
		my $num = $self->number_of_last;
		$num = shift if (@_);
		my $segments = $self->segments;
		return $$segments[$num]->paraxial_ray_a;
	}

	sub wavelength {
		my $self = shift;
		if (@_) {
			$self->first->{Wavelength} = shift;
		}
		$self->first->{Wavelength};
	}

	sub wavelength_name {
		my $self = shift;
		my $w = $self->wavelength;
		Wavelength->hash_key($w);
	}

	sub wavelength_value {
		my $self = shift;
		my $w = $self->wavelength;
		Wavelength->hash_value($w);
	}

	sub dump {
		my $self = shift;
		use Data::Dumper;
		my $dumper_data = Data::Dumper->Dump([$self],["RAY"]);
		print $dumper_data;
		return $self;
	}

	sub svg_path {
		my $self = shift;
		my $segments = $self->segments;
		return $self if (@$segments < 1);
		my $init_x = 100;
		$init_x = shift if (@_);
		my $init_y = 100;
		$init_y = shift if (@_);
		my $t = $$segments[0]->get_t;
		my $x0 = 0;
		my $s = 1;
		$s = -1 if ($$t[1] < 0);
		my $y0 = $s * ($$t[1]**2 + $$t[2]**2)**0.5;
		$init_y = $init_y + $y0;
		my $d = " M $init_x,$init_y m ";
		my $thickness = 0;
		my $switch = q{ l };
		foreach my $segment (@$segments) {
			next if ($segment->num_of_surface < 1);
			my $t = $segment->get_t;
			my $x = $$t[0] + $thickness - $x0;
			$s = -1;
			$s = 1 if ($$t[1] < 0);
			my $y =  $s * ($$t[1]**2 + $$t[2]**2)**0.5;
			my $y1 = $y - $y0;
			$thickness = $segment->thickness if ($segment->num_of_surface > 0);
			$d = $d."$x,$y1 \n";
			if ($switch) {
				$d = $d.$switch;
				$switch = q{};
			}
			$x0 = $$t[0];
			$y0 = $y;
		}
		return $d;
	}

	sub zero {
		my $value = shift;
		$value = shift if (ref $value);
		my $limit = 1e-14;
		#$limit = 1e-8;
		$value = 0 if ($value < $limit && $value > -$limit);
		return $value;
	}
	1;
}
