{ package Surface;
	# ONO Hiroki onohiroki@cup.com 
	# http://www.opensource.org/licenses/bsd-license
	# Copyright (c) 2011, ONO Hiroki onohiroki@cup.com 
	# All rights reserved.
	use strict;
	use warnings;
	sub new {
		my $class = shift;
		my $surface_number = 0;
		$surface_number = shift if (@_);
		my $radius = 0 ;
		$radius = shift if (@_);
		$radius = 0 if ($radius =~ m/Inf/i );
		my $curvature = 0;
		if ($radius eq q{}){
			$radius = 0;
			$curvature = 0;
		}
		elsif ($radius == 0) {
			$curvature = 0;
		}
		else {
			$curvature = 1/$radius;
		}
		my $thickness = 0;
		$thickness = shift if (@_);
		$thickness = 0 unless ($thickness);
		$thickness =~ s/Inf.*/0.1e14/i if ($thickness =~ m/Inf/i);
		my $glass = undef;
		$glass = shift if (@_);
		$glass = Air->new unless ($glass);
		my $stop = q{};
		$stop = shift if (@_);
		my $number_of_position = 1;
		my $self = {
			SurfaceNumber => $surface_number,
			Radius => $radius,
			Curvature => $curvature,
			Thickness => [$thickness],
			Glass => $glass,
			RefractMode => ['Refract'],
			Stop => $stop,
			NumberOfPosition => $number_of_position,
			Semidiameter => 0,
		}; 
		bless $self, $class;
	}

	sub new_with_power {
		my $class = shift;
		my $surface_number = shift;
		my $power = shift;
		my $thickness = shift;
		$thickness = 0 unless ($thickness);
		$thickness =~ s/Inf.*/0.1e14/i if ($thickness =~ m/Inf/i);
		my $glass = shift;
		$glass = Air->new unless ($glass);
		my $stop = q{};
		$stop = shift if (@_);
		#my $etc = \@_;
		#my $zoomed_thickness;
		#push (@$zoomed_thickness, $thickness);
		my $number_of_position = 1;
		my $self = {
			SurfaceNumber => $surface_number,
			Radius => 0,
			Curvature => 0,
			Power => $power,
			Thickness => [$thickness],
			Glass => $glass,
			RefractMode => ['Refract'],
			Stop => $stop,
			NumberOfPosition => $number_of_position,
			Semidiameter => 0,
		}; 
		bless $self, $class;
	}

	sub new_as_pupil {
		my $class = shift;
		my $thickness = shift;
		my $self = {
			SurfaceNumber => 0,
			Radius => 0,
			Curvature => 0,
			Thickness => [$thickness],
			Glass => Air->new,
			RefractMode => ['Refract'],
			Pupil => 'Pupil Surface',
		};
		bless $self, $class;
	}

	sub speak {
		my $self = shift;
		print "\nSurface ", $self->{SurfaceNumber};
		print ":";
		print "\n Surface number:    ", $self->number;
		print "\n Surface Type:      ", $self->type;
		print "\n Radius:            ", $self->radius;
		print "\n Thickness:         ", $self->thickness;
		print "\n Glass:             ", $self->glass_name;
		print "\n   Glass index:     ", $self->index;
		print "\n   Glass abbe num:  ", $self->abbe;
		print "\n Stop:              ", $self->stop;# if ($self->stop);
		print "\n Number of Position:", $self->number_of_position;
		print "\n--\n";
		return $self;
	}
	sub list {
		my $self = shift;
		my $test = q{};
		$test = shift if (@_);
		my $dummy_sur = shift if (@_);
		my $stop = $self->stop;
		my $format = "%7d%s ";
		my $sur_mark = ")";
		my $sur_number = $self->number;
		my $radius = $self->radius;
		my $thickness = $self->thickness;
		my $glass_name = $self->glass_name;
		my $glass_maker = $self->glass_maker;
		$glass_maker = q{} unless($glass_maker);
		my $glass_index = $self->index;
		my $glass_v = $self->abbe;
		my $refract_mode = q{};
		my $semidiameter = sprintf " %12s",q{};
		if ($self->reflect) {
			$refract_mode = "Reflect";
		}
		my $surface_mode = q{};

		if ($test =~ m/HEADER/ ) {
			return sprintf "%7s  %12s %12s %10s %-8s %9s %7s %12s", "Surface#", "Radius", "Thickness", "Glass", q{}, "Index", "V-Value", "Semidiameter";
		}
		else {
			if ($stop){
				$sur_mark = ">";
			}
			if ($test ne q{}) {
				$format = sprintf "%7s", $test;
				$format = "$format%s%s ";
				$sur_number = q{};
			}
			my $str_line_data = sprintf $format, $sur_number ,$sur_mark;
			foreach my $data ($radius, $thickness) {

				if ($data > 1e9 || $data < -1e9) {
					my $s = q{ };
					$s = '-' if ($data < -1);
					$format = ' %12s', ;
					$data = $s.'Infinity';
				}
				elsif ($data > 1e8 || $data < -1e8 || ($data < 0.0001 && $data > -0.0001 && $data != 0)){
					$format = ' %12.6e';
				}
				else{
					$format = ' %12.4f';
				}
				$str_line_data = $str_line_data.sprintf $format,$data;
			}
			my $str_glass;

			if ($dummy_sur && $dummy_sur eq 'dummy' && $refract_mode eq q{}){
				$format=" %10s %-8s %9s %7s";
				$str_glass = sprintf $format, q{}, q{}, q{}, q{};
			}
			elsif ($glass_index && $glass_index == 1 && $glass_v == 0 && $refract_mode eq q{}){
				$format=" %10s %-8s %9s %7s";
				$str_glass = sprintf $format, $glass_name, $glass_maker, q{}, q{};
			}
			elsif ($glass_index && $glass_v){
				$format=" %10s %-8s %9.6f %7.3f";
				$str_glass = sprintf $format, $glass_name, $glass_maker, $glass_index, $glass_v;
			}
			elsif ($glass_index){
				$format=" %10s %-8s %9.6f %7s";
				$str_glass = sprintf $format, $glass_name, $glass_maker, $glass_index, q{};
			}
			else {
				$format=" %10s %-8s %9s %7s";
				$str_glass = sprintf $format, $glass_name, $glass_maker, q{},q{};
			}

			$semidiameter = sprintf " %12.3f", $self->semidiameter if ($self->semidiameter);
			
			if (exists $self->{Power}) {
				$str_glass = sprintf "$str_glass    Power: %8.5f", $self->power;
			}
			if ($refract_mode) {
				$refract_mode = sprintf " %10s",$refract_mode;
			}
			if ((ref $self) ne "Surface") {
				$surface_mode = sprintf " %10s",(ref $self);
			}
			return $str_line_data.$str_glass.$refract_mode.$semidiameter.$surface_mode;
		}
	}
	sub number {
		my $self = shift;
		if (@_) {
			$self->{SurfaceNumber} = shift;
		}
		$self->{SurfaceNumber};
	}
	sub set_number {
		my $self = shift;
		my $num = shift;
		$self->{SurfaceNumber} = $num;
	}
	sub type {
		'Spherical';
	}
	sub radius {
		my $self = shift;
		$self->{Radius}
	}
	sub set_radius {
		my $self = shift;
		my $radius = shift;
		$self->{Radius} = $radius;
		if ($radius) {
			$self->{Curvature} = 1/$radius;
		}
		else {
			$self->{Curvature} = 0;
		}
	}
	sub curvature {
		my $self = shift;
		my $curvature = $self->{Curvature};
		$curvature = shift if (@_);
		$self->{Curvature} = $curvature;
		if ($curvature && $curvature != 0) {
			$self->{Radius} = 1/$curvature;
		}
		else {
			$self->{Radius} = 0;
		}
		$self->{Curvature};
	}

	sub set_curvature {
		my $self = shift;
		my $curvature = $self->{Curvature};
		$curvature = shift if (@_);
		$self->{Curvature} = $curvature;
		if ($curvature && $curvature != 0) {
			$self->{Radius} = 1/$curvature;
		}
		else {
			$self->{Radius} = 0;
		}
		return $self;
	}

	sub thickness {
		my $self = shift;
		my $position = 1;
		$position = shift if (@_);
		my $zoomed_thickness = $self->{Thickness};
		my $index = $position -1;
		$$zoomed_thickness[$index];
	}
	sub set_thickness {
		my $self = shift;
		my $thickness = shift;
		my $zoomed_thickness = $self->{Thickness};
		$$zoomed_thickness[0] = $thickness;
		return $self;
	}
	sub zooed_thickness {
		my $self = shift;
		unless (exists $self->{Thickness}) {
			$self->{Thickness} = [];
		}
		return $self->{Thickness};
	}
	sub set_zoomed_thickness {
		my $self = shift;
		my $zoomed_thickness = $self->zooed_thickness;
		@$zoomed_thickness = ();
		push(@$zoomed_thickness, @_);
		my $num_z = @$zoomed_thickness;
		$self->{NumberOfPosition} = $num_z;
		$zoomed_thickness;
	}
	sub set_thickness_with_position {
		my $self = shift;
		my $thickness = shift;
		my $position = shift;
		unless ($position) {
			print STDERR "ERROR:$self:set_thickness_with_position: number of argument";
		}
		my $zoomed_thickness = $self->zooed_thickness;
		if ($position > $self->number_of_position) {
			#$#{$zoomed_thickness} = $position;
		}
		
		$$zoomed_thickness[$position-1] = $thickness;
		my $number_of_pos = @$zoomed_thickness;
		$self->number_of_position($number_of_pos);
		$zoomed_thickness;
	}
	sub glass {
		my $self = shift;
		if (@_) {
			$self->{Glass} = shift;
		}
		$self->{Glass};
	}
	sub glass_name {
		my $self = shift;
		$self->glass->name;
	}
	sub glass_maker {
		my $self = shift;
		$self->glass->maker;
	}
	sub index {
		my $self = shift;
		my $index = 1;
		$index = $self->glass->index(@_);
		$index = -$index if ($self->reflect && $self->thickness < 0);
		return $index;
	}

	sub reflect {
		my $self = shift;
		my $mode = $self->{RefractMode};
		if ($$mode[$#{$mode}] && $$mode[$#{$mode}] eq 'Reflect'){
			return 'Reflect'
		}
		else {
			return undef;
		}
	}

	sub refract {
		my $self = shift;
		my $mode = $self->{RefractMode};
		if ($$mode[$#{$mode}] &&  $$mode[$#{$mode}] eq 'Refract'){
			return 'Refract'
		}
		else {
			return undef;
		}
	}

	sub get_refract_mode {
		my $self = shift;
		my $mode = $self->{RefractMode};
		return $$mode[$#{$mode}];
	}

	sub abbe {
		my $self = shift;
		$self->glass->abbe;
	}

	sub semidiameter {
		my $self = shift;
		$self->{Semidiameter} = shift if (@_);
		$self->{Semidiameter}
	}

	sub set_semidiameter {
		my $self = shift;
		$self->semidiameter(@_);
		return $self;
	}

	sub power {
		my $self = shift;
		my $n1 = shift;
		unless ($n1) {
			$n1 = 1;
		}
		if (exists $self->{Power}) {
			return $self->{Power};
		}
		my $n2 = $self->index(@_);
		my $r  = $self->radius;
		if ($r == 0) {
			0;
		}
		else {
			($n2-$n1)/$r; # 2.11
		}
	}

	sub set_stop {
		my $self = shift;
		if (@_) {
			$self->{Stop} = shift;
		}
		else {
			$self->{Stop} = 'stop';
		}
		return $self;
	}
	sub unset_stop {
		my $self = shift;
		$self->{Stop} = undef;
		return $self;
	}
	sub stop {
		my $self = shift;
		my $stop = $self->{Stop};
		unless ($stop) {
			return undef;
		}
		else {
			return $stop;
		}
	}

	sub set_refract_mode {
		my $self = shift;
		$self->{RefractMode} = \@_;
	}

	sub etc {
		my $self = shift;
		my $etc = $self->{etc};
		unless ($etc) {
			return q{};
		}
		else {
			return $etc;
		}
	}

	sub number_of_position {
		my $self = shift;
		$self->{NumberOfPosition} = shift if (@_);
		$self->{NumberOfPosition};
	}

	sub set_number_of_position {
		my $self = shift;
		$self->number_of_position(@_);
		return $self;
	}

	sub get_intersection {
		my $self = shift;
		my $ref_tvm1 = shift;
		my $ref_qv = shift;
		my $dvm1 = shift;

		#i)
		# pp 43
		my ($tvm1x, $tvm1y, $tvm1z)  = @$ref_tvm1; # Tv-1 (x~v-1)
		my ($qvx, $qvy, $qvz)        = @$ref_qv;  # Qv   (Xv)
		if ($dvm1 < 0) {
			$qvy = -$qvy;
			#s$mv = -$mv;
		}
		#pp 44
		my $r = $self->radius;

		my $pv = -(($tvm1x-$dvm1) * $qvx + $tvm1y * $qvy + $tvm1z * $qvz);
		my $mvx = ($tvm1x-$dvm1) + $pv * $qvx; # Mv   (Mxv)
		my $mv2 = ($tvm1x-$dvm1)**2 + $tvm1y**2 + $tvm1z**2 - $pv**2;
		my $qv;
		if ($r == 0 && Lens::zero($qvx) != 0) {
			$qv = $pv + (- 2 * $mvx) / ($qvx * ( 1 + 1));
		}
		elsif (Lens::zero($qvx) == 0) {
			print STDERR "\n",(ref $self),"::get_intersection, pv:$pv mvx:$mvx qvx:$qvx on Surface",$self->number,"\n";
			$qv = 1e14;
		}
		else {
			$qv = $pv + ($mv2 / $r - 2 * $mvx) / ($qvx * ( 1 + ( 1 - 1 / ($r * $qvx**2) * ($mv2 / $r - 2 * $mvx))**(1/2)));
		}

		# Tv   (x~v)
		my $tv0x = ($tvm1x-$dvm1) + $qv * $qvx;
		my $tv0y = $tvm1y + $qv * $qvy;
		my $tv0z = $tvm1z +  $qv * $qvz;

		my $lv;
		my $mv;
		my $nv;
		if ($r == 0) {
			$lv = 1;
			$mv = 0;
			$nv = 0;
		}
		else {
			$lv = 1 - $tv0x / $r;
			$mv = - $tv0y / $r;
			$nv = - $tv0z / $r;
		}
		#if ($dvm1 < 0) {
		#	print "\n********************",(ref $self),"::get_intersection d:$dvm1 m:$mv";
		#	#$qvy = -$qvy;
		#	#s$mv = -$mv;
		#}
		return [[$tv0x,$tv0y,$tv0z],[$lv, $mv, $nv],[$qvx, $qvy, $qvz]];
	}

	sub get_direction_cosines {
		my $self = shift;
		my $inv = shift;

		my $data = shift;
		shift @$data; # Tv   (x~v)
		my $ref_lmv= shift @$data; # (l, m, n)
		my $ref_qv = shift @$data; # Qv   (Xv)
		my $wavelength_name = undef;
		my $wavelength = shift if (@_);

		#pp 45
		#iii)
		my ($lv, $mv, $nv) = @$ref_lmv;
		my ($qvx, $qvy, $qvz)  = @$ref_qv;  # Qv   (Xv)
		my $inv1 = $self->index($wavelength);
		unless ($inv) {
			return undef;
		}
		my $ov = ($lv**2 + $mv**2 + $nv**2)**(1/2); # 3.19
		my $aoi = ($qvx * $lv + $qvy * $mv + $qvz * $nv)/$ov; # 3.20
		my $tr = (1  - ($inv / $inv1)**2 * (1 - $aoi**2))**(1/2); # 3.21
		if ($tr < 0) {
			print STDERR "\n",(ref $self),"::get_direction_cosines, Total Reflection on Surface",$self->number,".\n";
			#return [];
		}
		my $roi;
		if (Lens::zero($aoi) == 0) {
			$roi = $inv / abs($inv) * $inv1 / abs($inv1) * $tr;
		}
		else {
			$roi = $inv / abs($inv) * $inv1 / abs($inv1) * $aoi /abs($aoi) * $tr; # 3.21
		}
		
		my $gv = $roi - abs($inv / $inv1) * $aoi;
		my $qv1x = abs($inv / $inv1) * $qvx + $gv * $lv / $ov;
		my $qv1y = abs($inv / $inv1) * $qvy + $gv * $mv / $ov;
		my $qv1z = abs($inv / $inv1) * $qvz + $gv * $nv / $ov;
		return [$qv1x, $qv1y, $qv1z];
	}

	sub sag {
		my $self = shift;
		my $s1 = shift;
		my $s2 = 0;
		$s2 = shift if (@_);
		my $h2 = $s1**2 + $s2**2;
		my $c = $self->curvature;
		return $c * $h2 / (1 + (1 - $c**2 * $h2)**0.5);
	}

	sub svg_path {
		my $self = shift;
		my $r = $self->radius;
		my $r1r2;
		if ((ref $self->glass) ne 'Air') {
			$r1r2 = 'R1';
		}
		else {
			#print "\n",$self->number,$self->glass->name;
			$r1r2 = 'R2';
		}
		$r1r2 = shift if (@_);
		my $h = abs($r)/4;
		$h = shift if (@_);
		my $hh = $h * 2;
		my $sag;
		if ($r == 0) {
			$sag = 0;
		}
		else {
			$sag = $self->sag($h);
		}
		
		my $svg_path;
		my $arc = "0,0";
		my $sd = $self->semidiameter;
		$sd = $h unless ($sd);
	#	print STDERR "\n semidiameter:$sd h:$h";
		my $edge_sag = $sag;
		if ($r1r2 =~ m/R1/i) {
			if ($r == 0) {
				$svg_path = " m 0,-$h l 0,$hh ";
			}
			else {
				if ($r > 0) {
					$arc = "0,0";
				}
				else {
					$arc = "0,1";
					$r = -$r;
				}
				#if ($h > abs($r)) {
				if ($arc eq "0,1" && ($h > abs($r) || $h > $sd)) {
					$edge_sag = $self->sag($sd);
					my $a = $h - $sd;
					my $aa = $sd * 2;
					$svg_path = " m $edge_sag,-$h l 0,$a a$r,$r 0 $arc 0,$aa l 0,$a ";
					#$svg_path = " m $sag,-$h a$r,$r 0 $arc 0,$hh ";
				}
				else {
					$svg_path = " m $sag,-$h a$r,$r 0 $arc 0,$hh ";
				}
			}
		}
		else { # R2
			if ($r == 0) {
				$svg_path = " l 0,-$hh ";
				$edge_sag = 0;
			}
			else {
				if ($r > 0) {
					$arc = "0,1";
				}
				else {
					$arc = "0,0";
					$r = -$r;
				}
				if ($arc eq "0,1" && ($h > abs($r) || $h > $sd)) {
					
					$edge_sag = $self->sag($sd);
					my $a = $h - $sd;
					my $aa = $sd * 2;
					my $b = - $hh + $a;
					$svg_path = " 0,-$a a$r,$r 0 $arc 0,-$aa l 0,-$a ";
					#$svg_path = " a$r,$r 0 $arc 0,-$aa 0,$b";
				}
				else {
					$svg_path = " a$r,$r 0 $arc 0,-$hh ";
				}
			}
		}
		#print STDERR "\n           path:",$svg_path;
		return {SVGPath=>$svg_path,Sag=>$edge_sag};
	}

	sub dump {
		my $self = shift;
		use Data::Dumper;
		my $dumper_data = Data::Dumper->Dump([$self],["SURFACE"]);
		print $dumper_data;
		return $self;
	}

	sub change_as_aspherical {
		my $self = shift;
		my $new;
		my @keys = keys %$self;
		foreach (@keys) {
			$$new{$_} = $$self{$_};
		}
		my $asp = AsphericalSurface->copy_from_another_type($new);
		return $asp;
	}

	sub TO_JSON               
	{
		#http://d.hatena.ne.jp/koizuss/20090605/1244174060
		my $self = shift;
		my %h_self = map { $_ => $self->{$_} } keys(%$self);

		# インスタンスをhashに変換

		return { ref($self) => \%h_self };    # パッケージ名を付与して返却
	}

	sub bless_hash { #Surface
		my $class = shift;
		my $hash = shift;
		bless $hash, $class;
	}

	1;
}

{ package MatsuiAsphericalSurface;
	# ONO Hiroki onohiroki@cup.com 
	# open source project
	use strict;
	use warnings;
	@MatsuiAsphericalSurface::ISA = qw(AsphericalSurface);

	sub slope {
		my $self = shift;
		return undef;
	}
	
	sub list {
		my $self = shift;
		my $string = AsphericalSurface::list($self);
		my $aspheric_params = $self->aspheric_params;
		my $names;
		if ( $$aspheric_params[0] eq "Aspheric") {
			$aspheric_params = $self->aspheric_params_buff;
			$string = $string."\nType:Matsui\n        ";
			@$names = ('A'..'Z','AA'..'ZZ');
			my $num = 0;
			foreach (@$aspheric_params) {
				my $name = shift @$names;
				next unless ($_);
				$string = $string.sprintf "%5s:%20.10e",$name,$_ if ($_ != 0);
				if ($num == 2) {
					$string = $string."\n        ";
					$num = 0;
				}
				$num++;
			}
		}
		return $string;
	}

	sub power {
		my $self = shift;
		my $n1 = shift;
		unless ($n1) {
			$n1 = 1;
		}
		if (exists $self->{Power}) {
			return $self->{Power};
		}
		my $n2 = $self->index(@_);
		my $r  = $self->radius;
		my $asp = $self->aspheric_params_buff;
		if ($$asp[1]) {
			$r =  1 / ( (1 / $r) + 2 * $$asp[0]);
		}
		if ($r == 0) {
			0;
		}
		else {
			($n2-$n1)/$r;
		}
	}

	sub sag {
		my $self = shift;
		my $y = shift;
		my $z = shift;
		my $h2 = $y**2 + $z**2;
		my $r = $self->radius;
		my $sag = 0;
		my $asp = $self->aspheric_params;
		my $k = 0;
		if (@$asp > 1) {
			my @ac;
			push(@ac, @$asp);
			my $type = shift @ac;
			my $sum = 0;
			my $h_p = 1;
			foreach (@ac) {
				$h_p = $h_p * $h2;
				$sum = $sum + $_ * $h_p if ($_);
			}
			$sag = $r * ( 1 - ( 1 - $h2 / $r)**0.5 ) + $sum;
		}
		return $sag;
	}

	sub get_intersection {
		my $self = shift;
		my $step = Surface::get_intersection($self,@_);
		my $result = $self->get_intersection_on_aspheric($step);
		return $result;
	}

	sub get_intersection_on_aspheric {
		my $self = shift;
		my $step1 = shift;
		my $result = $step1;
		my $tv0 = $$step1[0];
		my $lmn = $$step1[1];
		my $ref_qv = $$step1[2];
		my ($tv0x,$tv0y,$tv0z) = @$tv0;
		my ($lv, $mv, $nv) = @$lmn;
		my $h2 = $tv0y**2 + $tv0z**2; # 3.12
		my $r = $self->radius; 
		my $r2 = $r**2;
		$lv = (1 - $h2 / $r2)**0.5; # 3.13
		my $xv00 = 0;
		my $asp;
		$asp = $self->aspheric_params_buff;
		if (@$asp > 1) {
			my @ac;
			push(@ac, @$asp);
			#my $type = shift @ac;
			my $sum = 0;
			my $h_p = 1;
			foreach (@ac) {
				$h_p = $h_p * $h2;
				$sum = $sum + $_ * $h_p if ($_);
				# pp45 3.14 AvHv**2 + BvHv**4 + ...
			}
			$xv00 = $r * ( 1 - $lv ) + $sum; # 3.14
			$sum = 0;
			$h_p = 1;
			my $d = 2;
			foreach (@ac) {
				$sum = $sum + $d * $_ * $h_p if ($_);
				$h_p = $h_p * $h2;
				$d = $d + 2;
				# pp45 3.15 2AvHv + 4BvHv**2 + ...
			}
			my $vv = 1 / $r + $lv * $sum; # 3.15
			$mv = - $tv0y * $vv; # 3.16
			$nv = - $tv0z * $vv; # 3.16
			my ($xx, $yy, $zz) = @$ref_qv;
			my $ev = $lv * ($xv00 - $tv0x) / ($xx * $lv + $yy * $mv + $zz * $nv); # 3.17
			my $xv1 = $tv0x + $ev * $xx; # 3.18
			my $yv1 = $tv0y + $ev * $yy; # 3.18
			my $zv1 = $tv0z + $ev * $zz; # 3.18
			my $test = abs($xv1 - $tv0x);
			($tv0x,$tv0y,$tv0z)=($xv1,$yv1,$zv1);
			if ($test eq 'nan'){
				$result = undef;
			}
			elsif (zero($test) == 0){
				$result = [[$tv0x,$tv0y,$tv0z],[$lv, $mv, $nv],$ref_qv];
			}
			else {
				$result = [[$tv0x,$tv0y,$tv0z],[$lv, $mv, $nv],$ref_qv];
				my $test = $self->get_intersection_on_aspheric([[$tv0x,$tv0y,$tv0z],[$lv, $mv, $nv],$ref_qv]);
				$result = $test if ($test);
			}
		}
		return $result;
	}
	1;
}

{ package AsphericalSurface;
	# ONO Hiroki onohiroki@cup.com 
	# open source project
	use strict;
	use warnings;
	@AsphericalSurface::ISA = qw(Surface);
	sub copy_from_another_type {
		my $class = shift;
		my $self = shift;
		$self->{AsphericConstant} = [] unless (exists $self->{AsphericConstant});
		$self->{AC_Buff} = [] unless (exists $self->{AC_Buff});
		$self->{AsphericConstantNames} = []unless (exists $self->{AsphericConstantNames});
		bless $self, $class;
	}
	sub type {
		'Aspherical';
	}

	sub set_aspheric_params {
		my $self = shift;
		$self->{AsphericConstant} = shift;
		return $self;
	}

	sub aspheric_params {
		my $self = shift;
		$self->{AsphericConstant} = shift if (@_);
		return $self->{AsphericConstant};
	}

	sub aspheric_params_names {
		my $self = shift;
		$self->{AsphericConstantNames} = shift if (@_);
		my $aspheric_params_names = $self->{AsphericConstantNames};
		if (@$aspheric_params_names < 1) {
			$self->set_aspheric_params_buff;
		}
		return $self->{AsphericConstantNames};
	}

	sub aspheric_params_buff {
		my $self = shift;
		$self->{AC_Buff} = shift if (@_);
		my $ac_buff = $self->{AC_Buff};
		if (@$ac_buff < 1) {
			$self->set_aspheric_params_buff;
		}
		return $self->{AC_Buff};
	}

	sub set_aspheric_params_buff {
		my $self = shift;
		$self->set_aspheric_params_buff_without_conic(@_);
	}

	sub set_aspheric_params_buff_without_conic {
		my $self = shift;
		my $ac = $self->aspheric_params;
		my $type = $$ac[0];
		my $ac_buf = [];
		
		my $names = [];
		my @l = ('A'..'Z','AA'..'ZZ');
		if ($type eq 'Matsui') {
			foreach (@$ac) {
				my $label = shift @l;
				push(@$names, $label);
			}
			push (@$ac_buf, @$ac);
			shift @$ac_buf;
		}
		elsif ($type eq 'Aspheric') {
			push(@$names, 'K');
			foreach (@$ac) {
				my $label = shift @l;
				push(@$names, $label);
			}
			# Malacara Handbook of optical design 2nd ed. (A2.8)
			$$ac[1] = 0 unless ($$ac[1]);
			$$ac[2] = 0 unless ($$ac[2]);
			$$ac[3] = 0 unless ($$ac[3]);
			$$ac[4] = 0 unless ($$ac[4]);
			$$ac[5] = 0 unless ($$ac[5]);
			my $k = $$ac[1];
			my $k1 = $k + 1;
			my $c = $self->curvature;
			my $d = 8;
			my $b1 = $$ac[2] + (  ($k1    - 1) * $c**3)/8;
			my $b2 = $$ac[3] + (  ($k1**2 - 1) * $c**5)/16 if 
			my $b3 = $$ac[4] + (5*($k1**3 - 1) * $c**7)/128;
			my $b4 = $$ac[5] + (7*($k1**4 - 1) * $c**9)/256;
			push (@$ac_buf, (0,$b1,$b2,$b3,$b4));
		}
		else {
			foreach (1..$#{$ac}) {
				push(@$names, "C$_");
			}
			push (@$ac_buf, @$ac);
			shift @$ac_buf;
		}
		$self->aspheric_params_buff($ac_buf);
		$self->aspheric_params_names($names);
		return $self;
	}

	sub list {
		my $self = shift;
		my $string = Surface::list($self);
		$string = sprintf "$string\n%8s", q{};
		my $aspheric_params = $self->aspheric_params;
		my $names = $self->aspheric_params_names;
		my @copy_of_names;
		push (@copy_of_names,'Type',@$names);
		my $num = 0;
		foreach (@$aspheric_params) {
			my $name = shift @copy_of_names;
			next unless ($_);
			unless ($name) {
				$string = $string.sprintf "_____:%20g",$_;
			}
			elsif ($name eq 'Type') {
				$string = $string.sprintf "%5s:%20s",$name,$_;
			}
			else {
				$string = $string.sprintf "%5s:%20.10e",$name,$_ if ($_);
			}
			
			if ($num == 2) {
				$string = $string."\n        ";
				$num = 0;
			}
			$num++;
		}
		return $string;
	}

	sub sag {
		my $self = shift;
		my $s = shift;
		my $sag = 0;
		my $asp = $self->aspheric_params;
		my $k = 0;
		if (@$asp > 1) {
			my @ac;
			push(@ac, @$asp);
			my $type = shift @ac;
			$k = shift @ac;
			my $c = $self->curvature;
			my $s2 = $s * $s;
			my $sum = 0;
			my $s_p = $s2;
			foreach (@ac) {
				$s_p = $s_p * $s2;
				$sum = $sum + $_ * $s_p if ($_);
			}
			$sag = $c * $s2 / (1 + (1 - ($k + 1) * $c**2 * $s2)**0.5 ) + $sum;
		}
		return $sag;
	}

	sub slope {
		my $self = shift;
		my $s = shift;
		my $sag = 0;
		my $asp = $self->aspheric_params;
		my $k = 0;
		if (@$asp > 1) {
			my @ac;
			push(@ac, @$asp);
			my $type = shift @ac;
			$k = shift @ac;
			my $c = $self->curvature;
			my $s2 = $s * $s;
			my $sum = 0;
			my $s_p = $s;
			my $p = 2;
			foreach (@ac) {
				$p = $p + 2;
				$s_p = $s_p * $s2;
				$sum = $sum + $p * $_ * $s_p if ($_);
			}
			$sag = $c * $s /  (1 - ($k + 1) * $c**2 * $s2)**0.5  + $sum;
		}
		return $sag;
	}

	sub get_intersection {
		my $self = shift;
		my $step = Surface::get_intersection($self,@_);
		my $result = $self->get_intersection_on_aspheric($step);
		return $result;
	}

	sub get_intersection_on_aspheric {
		my $self = shift;
		my $step1 = shift;
		my $depth = 0;
 		$depth = shift if (@_);
		my $result = $step1;
		my $tv0 = $$step1[0];
		my $lmn = $$step1[1];
		my $ref_qv = $$step1[2];
		my ($tv0x,$tv0y,$tv0z) = @$tv0;
		my ($lv, $mv, $nv) = @$lmn;
		my $h2 = $tv0y**2 + $tv0z**2; # matsui pp45 3.12
		my $r = $self->radius; 
		my $r2 = $r**2;
		my $xv00 = 0;
		my $asp;
		$asp = $self->aspheric_params; #
		if (@$asp > 1) {
			my @ac;
			push(@ac, @$asp);
			my $type = shift @ac;
			#print "\n",(ref $self),"::get_intersection_on_aspheric type:$type";
			my $k = shift @ac;
			my $aa = $k + 1;
			my $sum = 0;
			my $h_p = $h2;
			foreach (@ac) {
				$h_p = $h_p * $h2;
				$sum = $sum + $_ * $h_p if ($_);
				# kusakawa pp99 3.2.35
			}

			if ($aa != 0 && $r2 != 0) {
				$xv00 = $r / $aa - $r / $aa * (1 - $aa * ( $h2 / $r2 ))**0.5 + $sum;
			}
			else { # $aa = 0 or $r2 == 0
				$xv00 = $sum;
			}
			
			$sum = 0;
			$h_p = $h2;
			my $d = 2;
			foreach (@ac) {
				$sum = $sum + $d * $_ * $h_p if ($_);
				$h_p = $h_p * $h2;
				$d = $d + 2;
				# kusakawa pp99 3.2.38
			}
			my $bb = (1 - $aa * ($h2 / $r2))**0.5;
			$lv = 1;                                           # kusakawa pp99 3.2.37
			$mv = -1 * $tv0y / ($bb * $r ) - 2 * $tv0y * $sum; # kusakawa pp99 3.2.37
			$nv = -1 * $tv0z / ($bb * $r ) - 2 * $tv0z * $sum; # kusakawa pp99 3.2.37
			my ($xx, $yy, $zz) = @$ref_qv;
			my $ev = $lv * ($xv00 - $tv0x) / ($xx * $lv + $yy * $mv + $zz * $nv); # 3.17
			my $xv1 = $tv0x + $ev * $xx; # 3.18
			my $yv1 = $tv0y + $ev * $yy; # 3.18
			my $zv1 = $tv0z + $ev * $zz; # 3.18
			my $test = abs($xv1 - $tv0x);
			($tv0x,$tv0y,$tv0z)=($xv1,$yv1,$zv1);
			if ($test eq 'nan'){
				$result = undef;
			}
			elsif (Lens::zero($test) == 0){
				$result = [[$tv0x,$tv0y,$tv0z],[$lv, $mv, $nv],$ref_qv];
			}
			elsif ($depth > 5 && abs($test) < 1e-13){
				$result = [[$tv0x,$tv0y,$tv0z],[$lv, $mv, $nv],$ref_qv];
			}
			elsif ($depth > 20){
				print STDERR "\n",(ref $self),"::get_intersection_on_aspheric , Too Deep:$depth ! This function terminated. [$test]\n";
				$result = [[$tv0x,$tv0y,$tv0z],[$lv, $mv, $nv],$ref_qv];
			}
			else {
				$result = [[$tv0x,$tv0y,$tv0z],[$lv, $mv, $nv],$ref_qv];
				my $test = $self->get_intersection_on_aspheric([[$tv0x,$tv0y,$tv0z],[$lv, $mv, $nv],$ref_qv],++$depth);
				$result = $test if ($test);
			}
		}
		return $result;
	}

	sub change_as_matsui {
		my $self = shift;
		my $new;
		my @keys = keys %$self;
		foreach (@keys) {
			$$new{$_} = $$self{$_};
		}
		my $asp = MatsuiAsphericalSurface->copy_from_another_type($new);
		$self = $asp;
		return $asp;
	}
	1;
}

{ package SpecialSurface;
	# ONO Hiroki onohiroki@cup.com 
	# open source project
	use strict;
	use warnings;
	@SpecialSurface::ISA = qw(Surface);
	sub type {
		'Special';
	}

	1;
}


