{ package Wavelength;
	use strict;
	use warnings;
	# ONO Hiroki onohiroki@cup.com 
	# http://www.opensource.org/licenses/bsd-license
	# Copyright (c) 2011, ONO Hiroki onohiroki@cup.com 
	# All rights reserved.
	sub new {
		my $class = shift;
		my $self = {
			WavelengthBundle => Wavelength->default_wavelength_set->{WavelengthBundle},
			WavelengthArray => Wavelength->default_wavelength_set->{WavelengthArray},
		};
		bless $self, $class;	
	}

	sub default_wavelength_set {
		my $self=shift;
		my $names;
		@$names = qw/t s A' r C C' He-Ne D d e F F' He-Cd g h i/;
		my $wave_length = {
			q{t} => 1.01398,
			q{s} => 0.85211,
			q{A'} => 0.768195,
			q{r} => 0.706519,
			q{C} => 0.656273,
			q{C'} => 0.643847,
			q{He-Ne} => 0.632816,
			q{D} => 0.589294,
			q{d} => 0.587562,
			q{e} => 0.546074,
			q{F} => 0.486133,
			q{F'} => 0.479992,
			q{He-Cd} => 0.44157,
			q{g} => 0.435835,
			q{h} => 0.404656,
			q{i} => 0.365015,
		};
		my @keys = keys %$wave_length;
		foreach my $key (@keys) {
			$$wave_length{$key} = $$wave_length{$key} * 1000;
		}
		my $w = {
			WavelengthBundle => $wave_length,
			WavelengthArray => $names,
		};
		return $w;
	}

	sub wavelength_names {
		my $self = shift;
		my $name = undef;
		$name = shift if (@_);
		my $wave_length = Wavelength->default_wavelength_set->{WavelengthBundle};
		my @keys = keys %$wave_length;
		if ($name){
			if (exists $$wave_length{$name}) {
				return {$name=>$$wave_length{$name}};
			}
			elsif ($name =~ m/^[0-9]*\.?[0-9]*/) {
				my $result = {};
				foreach my $key (@keys) {
					my $test = $$wave_length{$key};
					$$result{$key} = $$wave_length{$key} if ($test =~ m/^$name/i);
				}
				return $result;
			}
		}
		return $wave_length;
	}

	sub new_single {
		my $class = shift;
		my $name = shift;
		my $wave_length_vale = shift;
		my $self = {
			WavelengthBundle => {$name => $wave_length_vale},
			WavelengthArray => [$name],
		};
		bless $self, $class;	
	}

	sub first {
		my $self = shift;
		my $na = $self->names_array;
		$self->by_wavelengths($$na[0]);
	}
	
	sub new_first {
		my $class = shift;
		my $wl = shift;
		my $self = {
			WavelengthBundle => $wl->bundle,
			WavelengthArray => $wl->names_array,
		};
		bless $self, $class;
		$self->first;
		return $self;
	}

	sub by_names {
		my $self=shift;
		my $names = $self->array;
		my $wave_length = $self->bundle;
		my $w;
		if (@_) {
			@$names = @_;
			foreach my $n (@$names) {
				if (exists $$wave_length{$n}) {
					$$w{$n}=$$wave_length{$n};
				}
				elsif ($n =~ m/^[0-9]*\.?[0-9]*$/) {
					my @keys = keys %$wave_length;
					foreach my $key (@keys) {
						$$w{$key} = $$wave_length{$key} if ($$wave_length{$key} =~ m/^$n/i);
					}
				}
			}
		}
		else {
			$w = $wave_length ;
		}
		$self->bundle($w);
		$self->array($names);
		return $self;
	}

	sub by_wavelengths {
		my $self=shift;
		my $names = $self->array;
		my $wavelength = $self->bundle;
		my $new_wl;
		my @new_names = ();
		if (@_) {
			my @keys = sort (keys %$wavelength);
			foreach (@_) {
				my $wl = $_;
				if ($wl =~ m/[A-Z]/i) {
					$wl = $self->wavelength($wl);
					next unless ($wl);
				}
				my $found = q{};
				foreach my $key (@keys) {
					if ( $$wavelength{$key} < $wl + 1 && $$wavelength{$key} > $wl -1 ) {
						$$new_wl{$key} = $$wavelength{$key}; 
						push (@new_names, $key);
						@keys = grep (!/^$key$/, @keys);
						$found = $key;
						last;
					}
				}
				unless ($found){
					my $nn = sprintf "%d",$wl;
					$$new_wl{$nn} = $wl;
					push (@new_names, $nn);
				}
			}
			$names = \@new_names;
		}
		else {
			$new_wl = $wavelength ;
		}
		$self->bundle($new_wl);
		$self->array($names);
		return $self;
	}
	
	sub copy_by_order {
		my $self=shift;
		my $names = $self->array;
		my $wavelength = $self->bundle;
		my $num = shift;
		my $wname = $$names[$num-1];
		my $wl = $$wavelength{$wname};
		my $new = Wavelength->new_single($wname,$wl);
		return $new;
	}

	sub get_order {
		my $self=shift;
		my $w = shift;
		my $names_arr = $self->names_array;
		my $order = $#{$names_arr};
		foreach ((reverse @$names_arr)) {
			last if ($_ eq $w->name);
			$order--;
		}
		$order++;
		return $order;
	}

	sub concat {
		my $self = shift;
		my $new = shift;
		my $self_buncle = $self->bundle;
		my $self_names = $self->array;
		my $new_bundle = $new->bundle;
		my $new_names = $new->array;
		foreach my $n (@$new_names) {
			$$self_buncle{$n} = $$new_bundle{$n};
		}
		$self->bundle($self_buncle);
		push(@$self_names,@$new_names);
		$self->array($self_names);
		return $self;
	}

	sub name {
		my $self = shift;
		my $names;
		unless (ref $self) {
			# class method
			$names = Wavelength->default_wavelength_set->{WavelengthArray};
		}
		else {
			$names = $self->array;
		}
		return $$names[0];
	}

	sub wavelength {
		my $self = shift;
		my $w;
		my $names;
		unless (ref $self) {
			# class method
			$w = Wavelength->default_wavelength_set->{WavelengthBundle};
			$names = Wavelength->default_wavelength_set->{WavelengthArray};
		}
		else {
			$w = $self->bundle;
			$names = $self->array;
		}
		if (@_) {
			my $name = shift;
			return $$w{$name};
		}
		else {
			return $$w{'d'} if (exists $$w{'d'});
			return $$w{$$names[0]};
		}
	}

	sub value {
		my $self = shift;
		$self->wavelength(@_);
	}

	sub hash_key {
		my $self = shift;
		my $hash_ref = shift;
		my @keys = keys %$hash_ref;
		return $keys[0];
	}

	sub hash_value {
		my $self = shift;
		my $hash_ref = shift;
		my @keys = keys %$hash_ref;
		if ($keys[0]) {
			return $$hash_ref{$keys[0]};
		}
		else {
			return undef;
		}
	}

	sub wavelength_array {
		my $self = shift;
		my $w;
		my $names;
		unless (ref $self) {
			# class method
			$w = Wavelength->default_wavelength_set->{WavelengthBundle};
			$names = Wavelength->default_wavelength_set->{WavelengthArray};
		}
		else {
			$w = $self->bundle;
			$names = $self->array;
		}
		my $a = [];
		foreach my $n (@$names) {
			push (@$a, $$w{$n});
		}
		return $a;
	}

	sub names_array {
		my $self = shift;
		$self->array(@_);
	}

	sub bundle {
		my $self = shift;
		if (@_) {
			$self->{WavelengthBundle} = shift;
		}
		$self->{WavelengthBundle};
	}

	sub array {
		my $self = shift;
		if (@_) {
			my $test = shift;
			$self->{WavelengthArray} = $test;
		}
		$self->{WavelengthArray};
	}

	sub list {
		my $self = shift;
		my $names = $self->names_array;
		my $w =  $self->bundle;
		printf "\n%11s", 'Wavelength';
		foreach (@$names) {
			printf " %8s", $_;
		}
		printf "\n%11s", q{};
		foreach (@$names) {
			printf " %8.3f", $$w{$_};
		}
		print "\n";
		return $self;
	}
	
	sub list_in_line {
		my $self = shift;
		my $names = $self->names_array;
		my $w =  $self->bundle;
		printf "%10s", 'Wavelength';
		foreach (@$names) {
			printf " %8s", $_;
		}
		foreach (@$names) {
			printf " %8.3f", $$w{$_};
		}
		return $self;
	}

	sub dump {
		my $self = shift;
		use Data::Dumper;
		my $dumper_data = Data::Dumper->Dump([$self],["WAVELENGTH"]);
		print $dumper_data;
		return $self;
	}
	1;
}
