{ package ZEMAX;
	# ONO Hiroki onohiroki@cup.com 
	# open source project
	use strict;
	use warnings;
	sub read_lens_data_zmx {
		my $data = shift;
		my @lines = split m{\n},$data;
		my $agf_glass_catalog;
		my $cv_glass_catalog;
		if (@_) {
			$agf_glass_catalog = shift;
			$cv_glass_catalog = shift;
		}
		if ((ref $agf_glass_catalog) ne 'GlassCagalog') {
			my $dumped;
			$dumped = `cat ~/Dropbox/work/perl-lens/Lib/GLASS/CVGLASS`;
			$cv_glass_catalog = GlassCatalog->new_with_dumped_data($dumped);
			$dumped = `cat ~/Dropbox/work/perl-lens/Lib/GLASS/AGFGLASS`;
			$agf_glass_catalog = GlassCatalog->new_with_dumped_data($dumped);
		}
		my $lens = Lens->new('lens');
		my $ref_wavelength = Wavelength->new->by_names('d');
		my $wavelength = $ref_wavelength;
		my @zoom_data = ();
		$lens->set_etc("Read ZEMAX .ZMX Lens Data");
		my %surface = (
			Number=>-1,
			Curvature=>0,
			Thickness=>0,
			Index=>1,
			Abbe=>0,
			Glass=>Air->get,
			RefractMode => 'Refract',
			Stop=>q{},
			AspericConstant=>[undef],
			);
		
		foreach (@lines) {
			s/\x0D|\x0A//gi; 
			$lens->set_etc($_) if (m/VERS\s+/);
			$lens->set_name($1) if (m/NAME\s+(.*)\s*$/);
			$lens->set_etc($_) if (m/UNIT\s+/);
			$lens->set_etc($_) if (m/ENPD\s+/);
			$lens->set_etc($_) if (m/REF\s+/);
			$lens->set_etc($_) if (m/GFAC\s+/);
			$lens->set_etc($_) if (m/FTYP\s+/);
			$lens->set_etc($_) if (m/XFLD\s+/);
			$lens->set_etc($_) if (m/YFLD\s+/);
			$lens->set_etc($_) if (m/FWGT\s+/);
			$lens->set_etc($_) if (m/WAVL\s+/);
			$lens->set_etc($_) if (m/WWGT\s+/);
			$lens->set_etc($_) if (m/PWAV\s+/);
			if (m/GLAS\s+/) {
				my $gline = $_;
				$gline =~ s/^\s*GLAS\s*//gi;
				my @g = split /\s+/, $gline;
				my $glass_name = shift @g;
				shift @g;
				shift @g;
				my $index = shift @g;
				my $abbe = shift @g;
				my $glass = $agf_glass_catalog->glass($glass_name);
				$glass = $cv_glass_catalog->glass($glass_name) if ((ref $glass) eq 'NotFound');
				#print "name:$glass_name i:$index v:$abbe\n";
				if ($glass_name eq 'MIRROR') {
					$surface{RefractMode} = 'Reflect';
					$surface{Glass} = Air->get;
				}
				elsif ((ref $glass) eq 'NotFound') {
					if ($glass_name =~ /[a-zA-Z0-9]/ && $index != 0 && $glass_name ne 'NotFound') {
						$surface{Glass} = Glass->new($glass_name,q{},$index,$abbe);
					}
					elsif ($index != 0) {
						$glass_name = "$index:$abbe";
						$surface{Glass} = Glass->new($glass_name,q{},$index,$abbe);
					}
				}
				else {
					$glass = $lens->glasscatalog->add_glass($glass);
					$surface{Glass} = $glass;
				}
			}
			if (m/WAVL\s+/) {
				my @wl = split /\s+/,$_;
				shift @wl;
				@wl = map { $_ * 1000 } @wl;
				$wavelength = Wavelength->new->by_wavelengths(@wl);
				$lens->wavelength($wavelength);
				#$wavelength->list;
			}
			if (m/PWAV\s+(.+)/){
				$ref_wavelength = $wavelength->copy_by_order($1);
			}
			if (m/\s*STOP\s*$/) {
				$surface{Stop} = 'yes';
			}
			$surface{Curvature} = $1 if (m/CURV\s+(\S+)/);
			$surface{Thickness} = $1 if (m/DISZ\s+(\S+)/);
			$surface{RefractMode} = 'Reflect' if (m/ REFL/);
			my $asp_const = $surface{AspericConstant};
			$$asp_const[1] = $1 if (m/^\s*CONI\s+([^;\s]+)/);
			$$asp_const[2] = $1 if (m/^\s*A\s+([^;\s]+)/);
			$$asp_const[3] = $1 if (m/^\s*B\s+([^;\s]+)/);
			$$asp_const[4] = $1 if (m/^\s*C\s+([^;\s]+)/);
			$$asp_const[5] = $1 if (m/^\s*D\s+([^;\s]+)/);
			$$asp_const[6] = $1 if (m/^\s*E\s+([^;\s]+)/);
			$$asp_const[7] = $1 if (m/^\s*F\s+([^;\s]+)/);
			$$asp_const[8] = $1 if (m/^\s*G\s+([^;\s]+)/);
			$$asp_const[9] = $1 if (m/^\s*H\s+([^;\s]+)/);
			$$asp_const[10] = $1 if (m/^\s*J\s+([^;\s]+)/);
			if ( m/SURF (.*)/ || m/^MNUM\s*$/ ) {
				my $sur_num = $1;
				#print "Glass:",(ref $surface{Glass}), ":";
				#print $surface{Glass}->name,"\n"; 
				if ($surface{Number} > -1){
					if ((ref $surface{Glass}) eq 'Glass') {
						$lens->add_surface(0,$surface{Thickness},$surface{Glass});
						$lens->last_surface->set_curvature($surface{Curvature});
					}
					else {
						$lens->add_surface(0,$surface{Thickness},Air->get);
						$lens->last_surface->set_curvature($surface{Curvature});
					}
					$lens->set_stop_surface($surface{Number}) if ($surface{Stop});
					$lens->surface->set_refract_mode('Reflect') if ($surface{RefractMode} eq 'Reflect');
					if (@$asp_const > 1) {
						if (@$asp_const < 2) {
							$$asp_const[0] = 'Conic';
						}
						else {
							$$asp_const[0] = 'Aspheric';
						}
						print join "\nASP:", @$asp_const;
						my $sur = $lens->last_surface->change_as_aspherical->set_aspheric_params($asp_const);
						my $num = $lens->last_surface->number;
						$lens->replace_surface($num,$sur);
						$surface{AspericConstant} = [undef];
						print "\n";
					}
				}
				$surface{RefractMode} = 'Refract';
				$surface{Stop} = q{};
				$surface{Glass} = Air->get;
				$surface{Number} = $sur_num;
			}

			if (m/THIC\s+(.*)/) {
				push(@zoom_data,$1);
			}
		}
		$lens->add_surface(0,$surface{Thickness},Air->get);
		if (@zoom_data) {
			my %z = {};
			foreach(@zoom_data){
				my @params = split(/\s+/,$_);
				$lens->set_thickness_with_position_of_surface($params[0],$params[2],$params[1]);
			}
		}
		$lens->wavelength($wavelength);
		$lens->reference_wavelength($ref_wavelength);
		$lens->set_indeces;
		return $lens;
	}
	sub read_lens_data_zmx_file {
		my $file = shift;
		my $lens;
		if ( -e $file ){
			open(IN, "<$file");
			my @lines = <IN>;
			my $data;
			$data = $data.$_ foreach (@lines);
			close(IN);
			$lens = read_lens_data_zmx($data);
		}
		else {
			print STDERR "\nERROR: can't find.\n";
			$lens = Lens->new;
		}
		return $lens;
	}
	
	sub read_glass_macro_data {
		my $self = shift;
		my $cv_glass_macro_data_catalog = GlassCatalog->new;
		my ($cat, $glass, $index, $vvalue) = (q{},q{},1,0);
		my @cv_glass_macro_data_files = <*.dat>;
		foreach my $dat (@cv_glass_macro_data_files) {
			open DATFILE, "<$dat";
			my $head = <DATFILE>;
			next if ($head !~ m/Catalog	Glass	Index	V-value	NF-NC	DPF	Price	Avail	Bubble	Stain/);
			while (<DATFILE>) {
				if (m/(\w+)\t([^\s]+)\t([^\s]+)\t([^\s]+)/) {
					$cat = $1;
					$glass = $2;
					$index = $3;
					$vvalue = $4;
					#printf "%10s %10s %10.6f 10.6f\n",$cat, $glass, $index, $vvalue;
					$cv_glass_macro_data_catalog->add_new_glass($glass,$cat,$index,$vvalue);
				}
			}
			close(DATFILE);
		}
		#$cv_glass_macro_data_catalog->list;
		delete($$cv_glass_macro_data_catalog{q{}});
		return $cv_glass_macro_data_catalog;
	}
	
	1;
}